# ruff: noqa: UP045
import math
import queue
import typing as ta

import anyio
import anyio.streams.memory

from ..queues import AsyncliteQueue
from ..queues import AsyncliteQueues
from .base import AnyioAsyncliteApi
from .base import AnyioAsyncliteObject


T = ta.TypeVar('T')


##


class AnyioAsyncliteQueue(AsyncliteQueue[T], AnyioAsyncliteObject):
    def __init__(
            self,
            send: anyio.streams.memory.MemoryObjectSendStream[T],
            recv: anyio.streams.memory.MemoryObjectReceiveStream[T],
            maxsize: int,
    ) -> None:
        super().__init__()

        self._send = send
        self._recv = recv
        self._maxsize = maxsize

    async def aclose(self) -> None:
        await self._send.aclose()
        await self._recv.aclose()

    def qsize(self) -> int:
        return self._recv.statistics().current_buffer_used

    def empty(self) -> bool:
        return self.qsize() == 0

    def full(self) -> bool:
        if self._maxsize <= 0:
            return False

        return self.qsize() >= self._maxsize

    async def put(self, item: T, *, timeout: ta.Optional[float] = None) -> None:
        if timeout is not None:
            with anyio.fail_after(timeout):
                await self._send.send(item)

        else:
            await self._send.send(item)

    def put_nowait(self, item: T) -> None:
        try:
            self._send.send_nowait(item)

        except anyio.WouldBlock as e:
            raise queue.Full from e

    async def get(self, *, timeout: ta.Optional[float] = None) -> T:
        if timeout is not None:
            with anyio.fail_after(timeout):
                return await self._recv.receive()

        else:
            return await self._recv.receive()

    def get_nowait(self) -> T:
        try:
            return self._recv.receive_nowait()

        except anyio.WouldBlock as e:
            raise queue.Empty from e


class AnyioAsyncliteQueues(AsyncliteQueues, AnyioAsyncliteApi):
    def make_queue(self, maxsize: int = 0) -> AsyncliteQueue:
        # anyio interprets 0 as "no buffer" (synchronous handoff), but stdlib queue uses 0 for unbounded. Convert 0 to
        # math.inf for anyio to get unbounded behavior.
        anyio_maxsize = math.inf if maxsize == 0 else maxsize
        send, recv = anyio.create_memory_object_stream(anyio_maxsize)
        return AnyioAsyncliteQueue(send, recv, maxsize)



import datetime


##


ISO_FMT = '%Y-%m-%dT%H:%M:%S%z'
ISO_FMT_US = '%Y-%m-%dT%H:%M:%S.%f%z'

ISO_FMT_NTZ = '%Y-%m-%dT%H:%M:%S'
ISO_FMT_US_NTZ = '%Y-%m-%dT%H:%M:%S.%f'


#


def utcnow() -> datetime.datetime:
    return datetime.datetime.now(tz=datetime.UTC)


def localnow() -> datetime.datetime:
    return datetime.datetime.now().astimezone()


def utcfromtimestamp(ts: float) -> datetime.datetime:
    return datetime.datetime.fromtimestamp(ts, tz=datetime.UTC)



# ruff: noqa: UP007 UP045
import abc
import functools
import operator
import typing as ta

from .abstract import Abstract


T = ta.TypeVar('T')
U = ta.TypeVar('U')


##


@functools.total_ordering
class Maybe(ta.Generic[T]):
    class ValueNotPresentError(BaseException):
        pass

    #

    @property
    @abc.abstractmethod
    def present(self) -> bool:
        raise NotImplementedError

    @abc.abstractmethod
    def must(self) -> T:
        raise NotImplementedError

    #

    @abc.abstractmethod
    def __repr__(self) -> str:
        raise NotImplementedError

    @abc.abstractmethod
    def __hash__(self) -> int:
        raise NotImplementedError

    @abc.abstractmethod
    def __eq__(self, other) -> bool:
        raise NotImplementedError

    @abc.abstractmethod
    def __lt__(self, other) -> bool:
        raise NotImplementedError

    #

    @ta.final
    def __ne__(self, other):
        return not (self == other)

    @ta.final
    def __iter__(self) -> ta.Iterator[T]:
        if self.present:
            yield self.must()

    @ta.final
    def __bool__(self) -> ta.NoReturn:
        raise TypeError

    #

    @ta.final
    def if_present(self, consumer: ta.Callable[[T], None]) -> None:
        if self.present:
            consumer(self.must())

    @ta.final
    def filter(self, predicate: ta.Callable[[T], bool]) -> 'Maybe[T]':
        if self.present and predicate(self.must()):
            return self
        else:
            return Maybe.empty()

    @ta.final
    def map(self, mapper: ta.Callable[[T], U]) -> 'Maybe[U]':
        if self.present:
            return Maybe.just(mapper(self.must()))
        else:
            return Maybe.empty()

    @ta.final
    def flat_map(self, mapper: ta.Callable[[T], 'Maybe[U]']) -> 'Maybe[U]':
        if self.present:
            if not isinstance(v := mapper(self.must()), Maybe):
                raise TypeError(v)
            return v
        else:
            return Maybe.empty()

    @ta.final
    def or_else(self, other: ta.Union[T, U]) -> ta.Union[T, U]:
        if self.present:
            return self.must()
        else:
            return other

    @ta.final
    def or_none(self) -> ta.Optional[T]:
        if self.present:
            return self.must()
        else:
            return None

    @ta.final
    def or_else_get(self, supplier: ta.Callable[[], ta.Union[T, U]]) -> ta.Union[T, U]:
        if self.present:
            return self.must()
        else:
            return supplier()

    @ta.final
    def or_else_raise(self, exception_supplier: ta.Callable[[], Exception]) -> T:
        if self.present:
            return self.must()
        else:
            raise exception_supplier()

    #

    @classmethod
    def of_optional(cls, v: ta.Optional[T]) -> 'Maybe[T]':
        if v is not None:
            return cls.just(v)
        else:
            return cls.empty()

    @classmethod
    def just(cls, v: T) -> 'Maybe[T]':
        return _JustMaybe(v)

    _empty: ta.ClassVar['Maybe']

    @classmethod
    def empty(cls) -> 'Maybe[T]':
        return Maybe._empty


##


class _Maybe(Maybe[T], Abstract):
    def __lt__(self, other):
        if not isinstance(other, _Maybe):
            return NotImplemented
        sp = self.present
        op = other.present
        if self.present and other.present:
            return self.must() < other.must()
        else:
            return op and not sp


@ta.final
class _JustMaybe(_Maybe[T]):
    __slots__ = ('_v', '_hash')

    def __init__(self, v: T) -> None:
        self._v = v

    @property
    def present(self) -> bool:
        return True

    def must(self) -> T:
        return self._v

    #

    def __repr__(self) -> str:
        return f'just({self._v!r})'

    _hash: int

    def __hash__(self) -> int:
        try:
            return self._hash
        except AttributeError:
            pass
        h = self._hash = hash((_JustMaybe, self._v))
        return h

    def __eq__(self, other):
        return (
            self.__class__ is other.__class__ and
            self._v == other._v  # noqa
        )


@ta.final
class _EmptyMaybe(_Maybe[T]):
    __slots__ = ()

    @property
    def present(self) -> bool:
        return False

    def must(self) -> T:
        raise Maybe.ValueNotPresentError

    #

    def __repr__(self) -> str:
        return 'empty()'

    def __hash__(self) -> int:
        return hash(_EmptyMaybe)

    def __eq__(self, other):
        return self.__class__ is other.__class__


Maybe._empty = _EmptyMaybe()  # noqa


##


setattr(Maybe, 'just', _JustMaybe)  # noqa
setattr(Maybe, 'empty', functools.partial(operator.attrgetter('_empty'), Maybe))



# ruff: noqa: UP006 UP007 UP045
# @omlish-lite
"""
TODO:
 - s/bytes/BytesLike?
"""
import io
import typing as ta

from ..lite.namespaces import NamespaceClass


##


class RawBytesReader(ta.Protocol):
    """Maps to `io.BufferedIOBase`."""

    def read1(self, n: int = -1, /) -> bytes:
        """Return at most `n` bytes with at most a single underlying read."""


class BytesReader(RawBytesReader, ta.Protocol):
    def read(self, n: int = -1, /) -> bytes:
        """Return exactly `n` bytes unless EOF is reached.."""


class BytesReaders(NamespaceClass):
    @ta.final
    class Empty:
        def read1(self, n: int = -1, /) -> bytes:
            return b''

        def read(self, n: int = -1, /) -> bytes:
            return b''

    @ta.final
    class Static:
        def __init__(self, b: bytes) -> None:
            self._r = io.BytesIO(b)

        def read1(self, n: int = -1, /) -> bytes:
            return self._r.read1(n)

        def read(self, n: int = -1, /) -> bytes:
            return self._r.read(n)

    @classmethod
    def of_bytes(cls, b: bytes) -> BytesReader:
        if b:
            return cls.Static(b)
        else:
            return cls.Empty()


##


class AsyncRawBytesReader(ta.Protocol):
    def read1(self, n: int = -1, /) -> ta.Awaitable[bytes]:
        """Return at most `n` bytes with at most a single underlying read."""


class AsyncBytesReader(AsyncRawBytesReader, ta.Protocol):
    def read(self, n: int = -1, /) -> ta.Awaitable[bytes]:
        """Return exactly `n` bytes unless EOF is reached.."""


class AsyncBytesReaders(NamespaceClass):
    @ta.final
    class Empty:
        async def read1(self, n: int = -1, /) -> bytes:
            return b''

        async def read(self, n: int = -1, /) -> bytes:
            return b''

    @ta.final
    class Static:
        def __init__(self, b: bytes) -> None:
            self._r = io.BytesIO(b)

        async def read1(self, n: int = -1, /) -> bytes:
            return self._r.read1(n)

        async def read(self, n: int = -1, /) -> bytes:
            return self._r.read(n)

    @classmethod
    def of_bytes(cls, b: bytes) -> AsyncBytesReader:
        if b:
            return cls.Static(b)
        else:
            return cls.Empty()



from .api import (  # noqa
    AnyioAsynclite as All,
)

from .base import (  # noqa
    AnyioAsyncliteObject as Object,
    AnyioAsyncliteApi as Api,
)

from .events import (  # noqa
    AnyioAsyncliteEvent as Event,
    AnyioAsyncliteEvents as Events,
)

from .identities import (  # noqa
    AnyioAsyncliteIdentities as Identities,
)

from .locks import (  # noqa
    AnyioAsyncliteLock as Lock,
    AnyioAsyncliteLocks as Locks,
)

from .queues import (  # noqa
    AnyioAsyncliteQueue as Queue,
    AnyioAsyncliteQueues as Queues,
)

from .semaphores import (  # noqa
    AnyioAsyncliteSemaphore as Semaphore,
    AnyioAsyncliteSemaphores as Semaphores,
)

from .sleeps import (  # noqa
    AnyioAsyncliteSleeps as Sleeps,
)



"""
TODO:
 - use lang.metadata?
"""
import typing as ta

from ... import check
from ... import collections as col
from ... import dataclasses as dc
from ... import lang
from ... import reflect as rfl
from ...lite import marshal as lm
from ..api.contexts import MarshalFactoryContext
from ..api.contexts import UnmarshalFactoryContext
from ..api.errors import UnhandledTypeError
from ..api.naming import Naming
from ..api.naming import translate_name
from ..api.options import Option
from ..api.types import Marshaler
from ..api.types import MarshalerFactory
from ..api.types import Unmarshaler
from ..api.types import UnmarshalerFactory
from .api import DEFAULT_FIELD_OPTIONS
from .api import DEFAULT_OBJECT_OPTIONS
from .api import FieldOptions
from .api import ObjectOptions
from .infos import FieldInfo
from .infos import FieldInfos
from .marshal import ObjectMarshaler
from .unmarshal import ObjectUnmarshaler


##


def get_dataclass_options(ty: type) -> ObjectOptions:
    return DEFAULT_OBJECT_OPTIONS.merge(*(dc.reflect(ty).spec.metadata_by_type.get(ObjectOptions, [])))


class _FieldInfoBuilder:
    def __init__(
            self,
            ty: type,
            opts: col.TypeMap[Option] | None = None,
    ) -> None:
        self.ty = ty
        if opts is None:
            opts = col.TypeMap()
        self.opts = opts

        self.obj_md = get_dataclass_options(ty)
        self.class_naming = self.obj_md.field_naming or opts.get(Naming)
        self.dc_rfl = dc.reflect(ty)
        self.type_hints = ta.get_type_hints(ty)

    def build_field_info(self, field: dc.Field) -> FieldInfo:
        """
        Merges configuration from multiple sources in this order (later = higher precedence):
        1. Empty baseline
        2. Class-level field_defaults (from ObjectMetadata)
        3. Field-level FieldMetadata (from field.metadata)
        4. Lite marshal compatibility overrides (OBJ_MARSHALER_FIELD_KEY, etc.)

        Then computes marshal/unmarshal names based on the merged configuration.
        """

        ##
        # Start with baseline (empty) and merge class-level defaults

        merged_md = DEFAULT_FIELD_OPTIONS.merge(self.obj_md.field_defaults)

        ##
        # Merge field-level FieldMetadata if present

        field_md = field.metadata.get(FieldOptions)
        if field_md is not None:
            merged_md = merged_md.merge(field_md)

        ##
        # Lite marshal compatibility - build override metadata

        lite_override_kw: dict[str, ta.Any] = {}

        # Handle OBJ_MARSHALER_FIELD_KEY
        if lm.OBJ_MARSHALER_FIELD_KEY in field.metadata:
            lfk = field.metadata[lm.OBJ_MARSHALER_FIELD_KEY]
            if lfk is not None:
                check.non_empty_str(lfk)
                lite_override_kw['name'] = lfk
            else:
                lite_override_kw['no_marshal'] = True
                lite_override_kw['no_unmarshal'] = True

        # Handle OBJ_MARSHALER_OMIT_IF_NONE
        if (lon := field.metadata.get(lm.OBJ_MARSHALER_OMIT_IF_NONE)) is not None:
            if check.isinstance(lon, bool):
                lite_override_kw['omit_if'] = lang.is_none

        # Merge lite overrides if any
        if lite_override_kw:
            merged_md = merged_md.merge(FieldOptions(**lite_override_kw))

        ##
        # Determine field type (with generic replacement if needed)

        if self.dc_rfl.spec.generic_init or merged_md.generic_replace:
            f_ty = rfl.to_annotation(self.dc_rfl.fields_inspection.generic_replaced_field_type(field.name))
        else:
            f_ty = self.type_hints[field.name]

        ##
        # Compute marshal/unmarshal names based on merged metadata

        has_explicit_name = merged_md.name is not None

        marshal_name: str | None
        unmarshal_names: ta.Sequence[str]

        if has_explicit_name:
            # Explicitly set name takes precedence
            # Type narrow: we know merged_md.name is not None here
            explicit_name = check.not_none(merged_md.name)
            marshal_name = explicit_name
            unmarshal_names = col.unique([explicit_name, *(merged_md.alts or ())])
        else:
            # Use naming convention if available, otherwise field name
            field_naming = field.metadata.get(Naming, self.class_naming)
            if field_naming is not None:
                base_name = translate_name(field.name, field_naming)
            else:
                base_name = field.name

            marshal_name = base_name
            unmarshal_names = [base_name]

        ##
        # Handle embed suffix (only if name wasn't explicitly set)

        if merged_md.embed and not has_explicit_name:
            # At this point marshal_name is guaranteed to be str (not None)
            marshal_name = check.not_none(marshal_name) + '_'
            unmarshal_names = [n + '_' for n in unmarshal_names]

        ##
        # Handle no_marshal/no_unmarshal

        if merged_md.no_marshal:
            marshal_name = None
        if merged_md.no_unmarshal:
            unmarshal_names = []

        ##
        # Create FieldInfo with computed values

        return FieldInfo(
            name=field.name,
            type=f_ty,
            marshal_name=marshal_name,
            unmarshal_names=unmarshal_names,
            options=merged_md,
        )

    def build_field_infos(self) -> FieldInfos:
        ret: list[FieldInfo] = []

        for field in self.dc_rfl.instance_fields:
            ret.append(self.build_field_info(field))

        return FieldInfos(ret)


def get_dataclass_field_infos(
        ty: type,
        opts: col.TypeMap[Option] | None = None,
) -> FieldInfos:
    return _FieldInfoBuilder(
        ty,
        opts,
    ).build_field_infos()


##


def _make_field_marshal_obj(
        ctx: MarshalFactoryContext,
        ty: ta.Any,
        obj: Marshaler | None,
        fac: MarshalerFactory | None,
        as_: ta.Any | None = None,
):
    if obj is not None:
        return obj
    if fac is not None:
        if (m := fac.make_marshaler(ctx, ty)) is None:
            raise UnhandledTypeError(ty)
        return m()
    if as_ is not None:
        return ctx.make_marshaler(as_)
    return ctx.make_marshaler(ty)


def _make_field_unmarshal_obj(
        ctx: UnmarshalFactoryContext,
        ty: ta.Any,
        obj: Unmarshaler | None,
        fac: UnmarshalerFactory | None,
        as_: ta.Any | None = None,
):
    if obj is not None:
        return obj
    if fac is not None:
        if (m := fac.make_unmarshaler(ctx, ty)) is None:
            raise UnhandledTypeError(ty)
        return m()
    if as_ is not None:
        return ctx.make_unmarshaler(as_)
    return ctx.make_unmarshaler(ty)


##


class AbstractDataclassFactory(lang.Abstract):
    def _get_options(self, ty: type) -> ObjectOptions:
        return get_dataclass_options(ty)

    def _get_field_infos(
            self,
            ty: type,
            opts: col.TypeMap[Option] | None = None,
    ) -> FieldInfos:
        return get_dataclass_field_infos(ty, opts)


##


def _type_or_generic_base(rty: rfl.Type) -> type | None:
    if isinstance(rty, rfl.Generic):
        return rty.cls
    elif isinstance(rty, type):
        return rty
    else:
        return None


class DataclassMarshalerFactory(AbstractDataclassFactory, MarshalerFactory):
    def make_marshaler(self, ctx: MarshalFactoryContext, rty: rfl.Type) -> ta.Callable[[], Marshaler] | None:
        if not (
            (ty := _type_or_generic_base(rty)) is not None and
            dc.is_dataclass(ty) and
            not lang.is_abstract_class(ty)
        ):
            return None

        def inner() -> Marshaler:
            ty = check.isinstance(_type_or_generic_base(rty), type)
            check.state(dc.is_dataclass(ty))
            check.state(not lang.is_abstract_class(ty))

            dc_md = self._get_options(ty)
            fis = self._get_field_infos(ty, ctx.options)

            fields = [
                (
                    fi,
                    _make_field_marshal_obj(
                        ctx,
                        fi.type,
                        fi.options.marshaler,
                        fi.options.marshaler_factory,
                        fi.options.marshal_as,
                    ),
                )
                for fi in fis
                if fi.name not in dc_md.specials.set
            ]

            return ObjectMarshaler(
                fields,
                specials=dc_md.specials,
            )

        return inner


##


class DataclassUnmarshalerFactory(AbstractDataclassFactory, UnmarshalerFactory):
    def make_unmarshaler(self, ctx: UnmarshalFactoryContext, rty: rfl.Type) -> ta.Callable[[], Unmarshaler] | None:
        if not (
            (ty := _type_or_generic_base(rty)) is not None and
            dc.is_dataclass(ty) and
            not lang.is_abstract_class(ty)
        ):
            return None

        def inner() -> Unmarshaler:
            ty = check.isinstance(_type_or_generic_base(rty), type)
            check.state(dc.is_dataclass(ty))
            check.state(not lang.is_abstract_class(ty))

            dc_md = self._get_options(ty)
            fis = self._get_field_infos(ty, ctx.options)

            d: dict[str, tuple[FieldInfo, Unmarshaler]] = {}
            defaults: dict[str, ta.Any] = {}
            embeds: dict[str, type] = {}
            embeds_by_unmarshal_name: dict[str, tuple[str, str]] = {}

            def add_field(fi: FieldInfo, *, prefixes: ta.Iterable[str] = ('',)) -> ta.Iterable[str]:
                ret: list[str] = []

                if fi.options.embed:
                    e_ty = check.isinstance(fi.type, type)
                    check.state(dc.is_dataclass(e_ty))
                    e_dc_md = get_dataclass_options(e_ty)
                    if e_dc_md.specials.set:
                        raise Exception(f'Embedded fields cannot have specials: {e_ty}')

                    embeds[fi.name] = e_ty
                    for e_fi in self._get_field_infos(e_ty, ctx.options):
                        e_ns = add_field(e_fi, prefixes=[p + ep for p in prefixes for ep in fi.unmarshal_names])
                        embeds_by_unmarshal_name.update({e_f: (fi.name, e_fi.name) for e_f in e_ns})
                        ret.extend(e_ns)

                else:
                    tup = (
                        fi,
                        _make_field_unmarshal_obj(
                            ctx,
                            fi.type,
                            fi.options.unmarshaler,
                            fi.options.unmarshaler_factory,
                            fi.options.unmarshal_as,
                        ),
                    )

                    for pfx in prefixes:
                        for un in fi.unmarshal_names:
                            un = pfx + un
                            if un in d:
                                raise KeyError(f'Duplicate fields for name {un!r}: {fi.name!r}, {d[un][0].name!r}')
                            d[un] = tup
                            ret.append(un)

                    if (dfl := fi.options.default) is not None and dfl.present:
                        defaults[fi.name] = dfl.must()

                return ret

            for fi in fis:
                if fi.name in dc_md.specials.set:
                    continue
                add_field(fi)

            return ObjectUnmarshaler(
                ty,
                d,
                specials=dc_md.specials,
                defaults=defaults,
                embeds=embeds,
                embeds_by_unmarshal_name=embeds_by_unmarshal_name,
                ignore_unknown=dc_md.ignore_unknown,
            )

        return inner



# ruff: noqa: UP006
# @omlish-lite
import abc
import typing as ta

from ....lite.abstract import Abstract
from ..core import IoPipelineHandler
from ..core import IoPipelineHandlerContext
from ..core import IoPipelineHandlerFn
from ..core import ShareableIoPipelineHandler
from ..flow.types import IoPipelineFlow
from ..flow.types import IoPipelineFlowMessages


##


class MessageToMessageDecoderIoPipelineHandler(IoPipelineHandler, Abstract):
    @abc.abstractmethod
    def _should_decode(self, ctx: IoPipelineHandlerContext, msg: ta.Any) -> bool:
        raise NotImplementedError

    @abc.abstractmethod
    def _decode(
            self,
            ctx: IoPipelineHandlerContext,
            msg: ta.Any,
            out: ta.List[ta.Any],
    ) -> None:
        raise NotImplementedError

    _called_decode = False
    _produced_messages = False

    def _on_inbound_flush_input(self, ctx: IoPipelineHandlerContext, msg: IoPipelineFlowMessages.FlushInput) -> None:  # noqa
        if not isinstance(self, ShareableIoPipelineHandler):
            if (
                    self._called_decode and
                    not self._produced_messages and
                    not ctx.services[IoPipelineFlow].is_auto_read()
            ):
                ctx.feed_out(IoPipelineFlowMessages.ReadyForInput())

            self._called_decode = False
            self._produced_messages = False

        ctx.feed_in(msg)

    def _on_inbound_should_decode(self, ctx: IoPipelineHandlerContext, msg: ta.Any) -> None:
        self._called_decode = True

        out: ta.List[ta.Any] = []

        self._decode(ctx, msg, out)

        if not out:
            return

        self._produced_messages = True

        for out_msg in out:
            ctx.feed_in(out_msg)

    def inbound(self, ctx: IoPipelineHandlerContext, msg: ta.Any) -> None:
        if isinstance(msg, IoPipelineFlowMessages.FlushInput):
            self._on_inbound_flush_input(ctx, msg)

        elif self._should_decode(ctx, msg):
            self._on_inbound_should_decode(ctx, msg)

        else:
            ctx.feed_in(msg)


##


class FnMessageToMessageDecoderIoPipelineHandler(MessageToMessageDecoderIoPipelineHandler):
    def __init__(
            self,
            filter_fn: IoPipelineHandlerFn[ta.Any, bool],  # noqa
            decode_fn: IoPipelineHandlerFn[ta.Any, ta.Iterable[ta.Any]],
    ) -> None:
        super().__init__()

        self._filter_fn = filter_fn
        self._decode_fn = decode_fn

    def _should_decode(self, ctx: IoPipelineHandlerContext, msg: ta.Any) -> bool:
        return self._filter_fn(ctx, msg)

    def _decode(
            self,
            ctx: IoPipelineHandlerContext,
            msg: ta.Any,
            out: ta.List[ta.Any],
    ) -> None:
        out.extend(self._decode_fn(ctx, msg))



# ruff: noqa: UP006 UP007 UP045
# @omlish-lite
import dataclasses as dc
import typing as ta

from ..core import IoPipelineMetadata


##


@dc.dataclass(frozen=True)
class DriverIoPipelineMetadata(IoPipelineMetadata):
    driver: ta.Any



import typing as ta

from .... import check
from .... import lang
from ..utils import SealableRegistry
from .base import Generator
from .base import Plan


K = ta.TypeVar('K')
V = ta.TypeVar('V')

GeneratorT = ta.TypeVar('GeneratorT', bound=Generator)


##


_GENERATOR_TYPES_BY_PLAN_TYPE: SealableRegistry[type[Plan], type[Generator]] = SealableRegistry()


def register_generator_type(pl_ty: type[Plan]) -> ta.Callable[[type[GeneratorT]], type[GeneratorT]]:
    def inner(g_ty):
        check.issubclass(g_ty, Generator)
        _GENERATOR_TYPES_BY_PLAN_TYPE[pl_ty] = g_ty
        return g_ty

    check.issubclass(pl_ty, Plan)
    return inner


@lang.cached_function
def generator_type_for_plan_type(pl_ty: type[Plan]) -> type[Generator]:
    return _GENERATOR_TYPES_BY_PLAN_TYPE[pl_ty]


@lang.cached_function
def all_generator_types() -> tuple[type[Generator], ...]:
    return tuple(sorted(
        {v for _, v in _GENERATOR_TYPES_BY_PLAN_TYPE.items()},
        key=lambda g_ty: g_ty.__name__,
    ))



import typing as ta

from ... import check
from ... import dataclasses as dc
from ... import lang
from ... import marshal as msh
from ... import reflect as rfl
from .. import jsonschema as jsch
from .openapi import Reference
from .openapi import Schema


##


def _reference_union_arg(rty: rfl.Type) -> rfl.Type | None:
    if isinstance(rty, rfl.Union) and len(rty.args) == 2 and Reference in rty.args:
        return check.single(a for a in rty.args if a is not Reference)
    else:
        return None


#


@dc.dataclass(frozen=True)
class _ReferenceUnionMarshaler(msh.Marshaler):
    m: msh.Marshaler
    r: msh.Marshaler

    def marshal(self, ctx: msh.MarshalContext, o: ta.Any) -> msh.Value:
        if isinstance(o, Reference):
            return self.r.marshal(ctx, o)
        else:
            return self.m.marshal(ctx, o)


class _ReferenceUnionMarshalerFactory(msh.MarshalerFactory):
    def make_marshaler(self, ctx: msh.MarshalFactoryContext, rty: rfl.Type) -> ta.Callable[[], msh.Marshaler] | None:
        if (rua := _reference_union_arg(rty)) is None:
            return None
        return lambda: _ReferenceUnionMarshaler(
            ctx.make_marshaler(check.not_none(rua)),
            ctx.make_marshaler(Reference),
        )


#


@dc.dataclass(frozen=True)
class _ReferenceUnionUnmarshaler(msh.Unmarshaler):
    u: msh.Unmarshaler
    r: msh.Unmarshaler

    def unmarshal(self, ctx: msh.UnmarshalContext, v: msh.Value) -> ta.Any:
        if not isinstance(v, ta.Mapping):
            raise TypeError(v)
        elif '$ref' in v:
            return self.r.unmarshal(ctx, v)
        else:
            return self.u.unmarshal(ctx, v)  # noqa


class _ReferenceUnionUnmarshalerFactory(msh.UnmarshalerFactory):
    def make_unmarshaler(self, ctx: msh.UnmarshalFactoryContext, rty: rfl.Type) -> ta.Callable[[], msh.Unmarshaler] | None:  # noqa
        if (rua := _reference_union_arg(rty)) is None:
            return None
        return lambda: _ReferenceUnionUnmarshaler(
            ctx.make_unmarshaler(check.not_none(rua)),
            ctx.make_unmarshaler(Reference),
        )


##


@dc.dataclass(frozen=True)
class _SchemaMarshaler(msh.Marshaler):
    m_dct: ta.Mapping[str, tuple[str, msh.Marshaler]]
    kw_m: msh.Marshaler

    def marshal(self, ctx: msh.MarshalContext, o: ta.Any) -> msh.Value:
        sch: Schema = check.isinstance(o, Schema)
        dct = {}
        for f, (k, fm) in self.m_dct.items():
            fv = getattr(sch, f)
            if fv is None:
                continue
            dct[k] = fm.marshal(ctx, fv)
        if sch.keywords is not None:
            dct.update(keywords=self.kw_m.marshal(ctx, sch.keywords))
        return dct


class _SchemaMarshalerFactory(msh.MarshalerFactory):
    def make_marshaler(self, ctx: msh.MarshalFactoryContext, rty: rfl.Type) -> ta.Callable[[], msh.Marshaler] | None:
        if rty is not Schema:
            return None
        return lambda: _SchemaMarshaler(
            {
                f: (msh.translate_name(f, msh.Naming.LOW_CAMEL), ctx.make_marshaler(rfl.type_(a)))
                for f, a in dc.reflect(Schema).field_annotations.items()
                if f != 'keywords'
            },
            ctx.make_marshaler(jsch.Keywords),
        )


#


@dc.dataclass(frozen=True)
class _SchemaUnmarshaler(msh.Unmarshaler):
    u_dct: ta.Mapping[str, tuple[str, msh.Unmarshaler]]
    kw_u: msh.Unmarshaler

    def unmarshal(self, ctx: msh.UnmarshalContext, v: msh.Value) -> ta.Any:
        dct = dict(check.isinstance(v, ta.Mapping))
        kw = {}
        for k, (f, fu) in self.u_dct.items():
            try:
                kv = dct.pop(k)
            except KeyError:
                continue
            kw[f] = fu.unmarshal(ctx, kv)
        if dct:
            kw.update(keywords=self.kw_u.unmarshal(ctx, dct))
        return Schema(**kw)


class _SchemaUnmarshalerFactory(msh.UnmarshalerFactory):
    def make_unmarshaler(self, ctx: msh.UnmarshalFactoryContext, rty: rfl.Type) -> ta.Callable[[], msh.Unmarshaler] | None:  # noqa
        if rty is not Schema:
            return None
        return lambda: _SchemaUnmarshaler(
            {
                msh.translate_name(f, msh.Naming.LOW_CAMEL): (f, ctx.make_unmarshaler(rfl.type_(a)))
                for f, a in dc.reflect(Schema).field_annotations.items()
                if f != 'keywords'
            },
            ctx.make_unmarshaler(jsch.Keywords),
        )


##


@lang.static_init
def _install_standard_marshaling() -> None:
    msh.install_standard_factories(
        _ReferenceUnionMarshalerFactory(),
        _ReferenceUnionUnmarshalerFactory(),

        _SchemaMarshalerFactory(),
        _SchemaUnmarshalerFactory(),
    )



# ruff: noqa: UP006 UP007 UP043 UP045
# @omlish-lite
##
# MIT License
#
# Copyright (c) 2019 Masaaki Goshima
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
# persons to whom the Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
# OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
##
import abc
import dataclasses as dc
import enum
import io
import typing as ta
import unicodedata

from ....lite.abstract import Abstract
from ....lite.check import check
from ....lite.dataclasses import dataclass_field_required
from .errors import EofYamlError
from .errors import YamlError
from .errors import YamlErrorOr
from .errors import yaml_error
from .tokens import YamlNumberType
from .tokens import YamlPosition
from .tokens import YamlToken
from .tokens import YamlTokenType
from .tokens import yaml_detect_line_break_char
from .tokens import yaml_literal_block_header
from .tokens import yaml_to_number


##


@dc.dataclass()
class UnexpectedNodeTypeYamlError(YamlError):
    actual: 'YamlNodeType'
    expected: 'YamlNodeType'
    token: YamlToken

    @property
    def message(self) -> str:
        return f'unexpected node type: expected {self.expected.name}, got {self.actual.name}, at {self.token.position}'


##


class YamlAstErrors:
    def __new__(cls, *args, **kwargs):  # noqa
        raise TypeError

    INVALID_TOKEN_TYPE = yaml_error('invalid token type')
    INVALID_ANCHOR_NAME = yaml_error('invalid anchor name')
    INVALID_ALIAS_NAME = yaml_error('invalid alias name')


class YamlNodeType(enum.Enum):
    # UNKNOWN type identifier for default
    UNKNOWN = enum.auto()
    # DOCUMENT type identifier for document node
    DOCUMENT = enum.auto()
    # NULL type identifier for null node
    NULL = enum.auto()
    # BOOL type identifier for boolean node
    BOOL = enum.auto()
    # INTEGER type identifier for integer node
    INTEGER = enum.auto()
    # FLOAT type identifier for float node
    FLOAT = enum.auto()
    # INFINITY type identifier for infinity node
    INFINITY = enum.auto()
    # NAN type identifier for nan node
    NAN = enum.auto()
    # STRING type identifier for string node
    STRING = enum.auto()
    # MERGE_KEY type identifier for merge key node
    MERGE_KEY = enum.auto()
    # LITERAL type identifier for literal node
    LITERAL = enum.auto()
    # MAPPING type identifier for mapping node
    MAPPING = enum.auto()
    # MAPPING_KEY type identifier for mapping key node
    MAPPING_KEY = enum.auto()
    # MAPPING_VALUE type identifier for mapping value node
    MAPPING_VALUE = enum.auto()
    # SEQUENCE type identifier for sequence node
    SEQUENCE = enum.auto()
    # SEQUENCE_ENTRY type identifier for sequence entry node
    SEQUENCE_ENTRY = enum.auto()
    # ANCHOR type identifier for anchor node
    ANCHOR = enum.auto()
    # ALIAS type identifier for alias node
    ALIAS = enum.auto()
    # DIRECTIVE type identifier for directive node
    DIRECTIVE = enum.auto()
    # TAG type identifier for tag node
    TAG = enum.auto()
    # COMMENT type identifier for comment node
    COMMENT = enum.auto()
    # COMMENT_GROUP type identifier for comment group node
    COMMENT_GROUP = enum.auto()


# String node type identifier to YAML Structure name based on https://yaml.org/spec/1.2/spec.html
YAML_NODE_TYPE_YAML_NAMES: ta.Mapping[YamlNodeType, str] = {
    YamlNodeType.UNKNOWN: 'unknown',
    YamlNodeType.DOCUMENT: 'document',
    YamlNodeType.NULL: 'null',
    YamlNodeType.BOOL: 'boolean',
    YamlNodeType.INTEGER: 'int',
    YamlNodeType.FLOAT: 'float',
    YamlNodeType.INFINITY: 'inf',
    YamlNodeType.NAN: 'nan',
    YamlNodeType.STRING: 'string',
    YamlNodeType.MERGE_KEY: 'merge key',
    YamlNodeType.LITERAL: 'scalar',
    YamlNodeType.MAPPING: 'mapping',
    YamlNodeType.MAPPING_KEY: 'key',
    YamlNodeType.MAPPING_VALUE: 'value',
    YamlNodeType.SEQUENCE: 'sequence',
    YamlNodeType.SEQUENCE_ENTRY: 'value',
    YamlNodeType.ANCHOR: 'anchor',
    YamlNodeType.ALIAS: 'alias',
    YamlNodeType.DIRECTIVE: 'directive',
    YamlNodeType.TAG: 'tag',
    YamlNodeType.COMMENT: 'comment',
    YamlNodeType.COMMENT_GROUP: 'comment',
}


##


# Node type of node
class YamlNode(Abstract):
    # io.Reader

    def __str__(self) -> ta.NoReturn:
        raise TypeError

    @abc.abstractmethod
    def string(self) -> str:
        # FIXME: migrate off - ensure all sprintfy things explicitly call .string()
        raise NotImplementedError

    # get_token returns token instance
    @abc.abstractmethod
    def get_token(self) -> ta.Optional[YamlToken]:
        raise NotImplementedError

    # type returns type of node
    @abc.abstractmethod
    def type(self) -> YamlNodeType:
        raise NotImplementedError

    # add_column add column number to child nodes recursively
    @abc.abstractmethod
    def add_column(self, column: int) -> None:
        raise NotImplementedError

    # set_comment set comment token to node
    @abc.abstractmethod
    def set_comment(self, node: ta.Optional['CommentGroupYamlNode']) -> ta.Optional[YamlError]:
        raise NotImplementedError

    # comment returns comment token instance
    @abc.abstractmethod
    def get_comment(self) -> ta.Optional['CommentGroupYamlNode']:
        raise NotImplementedError

    # get_path returns YAMLPath for the current node
    @abc.abstractmethod
    def get_path(self) -> str:
        raise NotImplementedError

    # set_path set YAMLPath for the current node
    @abc.abstractmethod
    def set_path(self, path: str) -> None:
        raise NotImplementedError

    # marshal_yaml
    @abc.abstractmethod
    def marshal_yaml(self) -> YamlErrorOr[str]:
        raise NotImplementedError

    # already read length
    @abc.abstractmethod
    def read_len(self) -> int:
        raise NotImplementedError

    # append read length
    @abc.abstractmethod
    def append_read_len(self, n: int) -> None:
        raise NotImplementedError

    # clean read length
    @abc.abstractmethod
    def clear_len(self) -> None:
        raise NotImplementedError


# MapKeyNode type for map key node
class MapKeyYamlNode(YamlNode, Abstract):
    @abc.abstractmethod
    def is_merge_key(self) -> bool:
        raise NotImplementedError

    # String node to text without comment
    @abc.abstractmethod
    def string_without_comment(self) -> str:
        raise NotImplementedError


# ScalarNode type for scalar node
class ScalarYamlNode(MapKeyYamlNode, Abstract):
    @abc.abstractmethod
    def get_value(self) -> ta.Any:
        raise NotImplementedError


##


@dc.dataclass()
class BaseYamlNode(YamlNode, Abstract):
    path: str = ''
    comment: ta.Optional['CommentGroupYamlNode'] = None
    cur_read: int = 0

    def read_len(self) -> int:
        return self.cur_read

    def clear_len(self) -> None:
        self.cur_read = 0

    def append_read_len(self, l: int) -> None:
        self.cur_read += l

    # get_path returns YAMLPath for the current node.
    @ta.final
    def get_path(self: ta.Optional['BaseYamlNode']) -> str:
        if self is None:
            return ''
        return self.path

    # set_path set YAMLPath for the current node.
    @ta.final
    def set_path(self: ta.Optional['BaseYamlNode'], path: str) -> None:
        if self is None:
            return
        self.path = path

    # get_comment returns comment token instance
    def get_comment(self) -> ta.Optional['CommentGroupYamlNode']:
        return self.comment

    # set_comment set comment token
    def set_comment(self, node: ta.Optional['CommentGroupYamlNode']) -> ta.Optional[YamlError]:
        self.comment = node
        return None


def yaml_add_comment_string(base: str, node: 'CommentGroupYamlNode') -> str:
    return f'{base} {node.string()}'


##


def yaml_read_node(p: str, node: YamlNode) -> YamlErrorOr[int]:
    s = node.string()
    read_len = node.read_len()
    remain = len(s) - read_len
    if remain == 0:
        node.clear_len()
        return EofYamlError()

    size = min(remain, len(p))
    for idx, b in enumerate(s[read_len:read_len + size]):
        p[idx] = b  # type: ignore[index]  # FIXME: lol

    node.append_read_len(size)
    return size


def yaml_check_line_break(t: YamlToken) -> bool:
    if t.prev is not None:
        lbc = '\n'
        prev = t.prev
        adjustment = 0
        # if the previous type is sequence entry use the previous type for that
        if prev.type == YamlTokenType.SEQUENCE_ENTRY:
            # as well as switching to previous type count any new lines in origin to account for:
            # -
            #   b: c
            adjustment = t.origin.rstrip(lbc).count(lbc)
            if prev.prev is not None:
                prev = prev.prev

        line_diff = t.position.line - prev.position.line - 1
        if line_diff > 0:
            if prev.type == YamlTokenType.STRING:
                # Remove any line breaks included in multiline string
                adjustment += prev.origin.strip().rstrip(lbc).count(lbc)

            # Due to the way that comment parsing works its assumed that when a null value does not have new line in
            # origin it was squashed therefore difference is ignored.
            # foo:
            #  bar:
            #  # comment
            #  baz: 1
            # becomes
            # foo:
            #  bar: null # comment
            #
            #  baz: 1
            if prev.type in (YamlTokenType.NULL, YamlTokenType.IMPLICIT_NULL):
                return prev.origin.count(lbc) > 0

            if line_diff - adjustment > 0:
                return True

    return False


##


class YamlAsts:
    # Null create node for null value
    @classmethod
    def null(cls, tk: YamlToken) -> 'NullYamlNode':
        return NullYamlNode(
            token=tk,
        )

    _BOOL_TRUE_STRS: ta.ClassVar[ta.AbstractSet[str]] = {'1', 't', 'T', 'true', 'TRUE', 'True'}
    _BOOL_FALSE_STRS: ta.ClassVar[ta.AbstractSet[str]] = {'0', 'f', 'F', 'false', 'FALSE', 'False'}

    @classmethod
    def _parse_bool(cls, s: str) -> bool:
        if s in cls._BOOL_TRUE_STRS:
            return True
        if s in cls._BOOL_FALSE_STRS:
            return False
        raise ValueError(f'"{s}" is not a valid boolean string')

    # bool_ create node for boolean value
    @classmethod
    def bool_(cls, tk: YamlToken) -> 'BoolYamlNode':
        b = cls._parse_bool(tk.value)
        return BoolYamlNode(
            token=tk,
            value=b,
        )

    # integer create node for integer value
    @classmethod
    def integer(cls, tk: YamlToken) -> 'IntegerYamlNode':
        v: ta.Any = None
        if (num := yaml_to_number(tk.value)) is not None:
            v = num.value

        return IntegerYamlNode(
            token=tk,
            value=v,
        )

    # float_ create node for float value
    @classmethod
    def float_(cls, tk: YamlToken) -> 'FloatYamlNode':
        v: float = 0.
        if (num := yaml_to_number(tk.value)) is not None and num.type == YamlNumberType.FLOAT:
            if isinstance(num.value, float):
                v = num.value

        return FloatYamlNode(
            token=tk,
            value=v,
        )

    # infinity create node for .inf or -.inf value
    @classmethod
    def infinity(cls, tk: YamlToken) -> 'InfinityYamlNode':
        if tk.value in ('.inf', '.Inf', '.INF'):
            value = float('inf')
        elif tk.value in ('-.inf', '-.Inf', '-.INF'):
            value = float('-inf')
        node = InfinityYamlNode(
            token=tk,
            value=value,
        )
        return node

    # nan create node for .nan value
    @classmethod
    def nan(cls, tk: YamlToken) -> 'NanYamlNode':
        return NanYamlNode(
            token=tk,
        )

    # string create node for string value
    @classmethod
    def string(cls, tk: YamlToken) -> 'StringYamlNode':
        return StringYamlNode(
            token=tk,
            value=tk.value,
        )

    # comment create node for comment
    @classmethod
    def comment(cls, tk: ta.Optional[YamlToken]) -> 'CommentYamlNode':
        return CommentYamlNode(
            token=tk,
        )

    @classmethod
    def comment_group(cls, comments: ta.Iterable[ta.Optional[YamlToken]]) -> 'CommentGroupYamlNode':
        nodes: ta.List[CommentYamlNode] = []
        for c in comments:
            nodes.append(cls.comment(c))

        return CommentGroupYamlNode(
            comments=nodes,
        )

    # merge_key create node for merge key ( << )
    @classmethod
    def merge_key(cls, tk: YamlToken) -> 'MergeKeyYamlNode':
        return MergeKeyYamlNode(
            token=tk,
        )

    # mapping create node for map
    @classmethod
    def mapping(cls, tk: YamlToken, is_flow_style: bool, *values: 'MappingValueYamlNode') -> 'MappingYamlNode':
        node = MappingYamlNode(
            start=tk,
            is_flow_style=is_flow_style,
            values=[],
        )
        node.values.extend(values)
        return node

    # mapping_value create node for mapping value
    @classmethod
    def mapping_value(cls, tk: YamlToken, key: 'MapKeyYamlNode', value: YamlNode) -> 'MappingValueYamlNode':
        return MappingValueYamlNode(
            start=tk,
            key=key,
            value=value,
        )

    # mapping_key create node for map key ( '?' ).
    @classmethod
    def mapping_key(cls, tk: YamlToken) -> 'MappingKeyYamlNode':
        return MappingKeyYamlNode(
            start=tk,
        )

    # sequence create node for sequence
    @classmethod
    def sequence(cls, tk: YamlToken, is_flow_style: bool) -> 'SequenceYamlNode':
        return SequenceYamlNode(
            start=tk,
            is_flow_style=is_flow_style,
            values=[],
        )

    @classmethod
    def anchor(cls, tk: YamlToken) -> 'AnchorYamlNode':
        return AnchorYamlNode(
            start=tk,
        )

    @classmethod
    def alias(cls, tk: YamlToken) -> 'AliasYamlNode':
        return AliasYamlNode(
            start=tk,
        )

    @classmethod
    def document(cls, tk: ta.Optional[YamlToken], body: ta.Optional[YamlNode]) -> 'DocumentYamlNode':
        return DocumentYamlNode(
            start=tk,
            body=body,
        )

    @classmethod
    def directive(cls, tk: YamlToken) -> 'DirectiveYamlNode':
        return DirectiveYamlNode(
            start=tk,
        )

    @classmethod
    def literal(cls, tk: YamlToken) -> 'LiteralYamlNode':
        return LiteralYamlNode(
            start=tk,
        )

    @classmethod
    def tag(cls, tk: YamlToken) -> 'TagYamlNode':
        return TagYamlNode(
            start=tk,
        )


##


# File contains all documents in YAML file
@dc.dataclass()
class YamlFile:
    name: str = ''
    docs: ta.List['DocumentYamlNode'] = dc.field(default_factory=list)

    # read implements (io.Reader).Read
    def read(self, p: str) -> YamlErrorOr[int]:
        for doc in self.docs:
            n = doc.read(p)
            if isinstance(n, EofYamlError):
                continue
            return n
        return EofYamlError()

    # string all documents to text
    def string(self) -> str:
        docs: ta.List[str] = []
        for doc in self.docs:
            docs.append(doc.string())
        if len(docs) > 0:
            return '\n'.join(docs) + '\n'
        else:
            return ''


##


# DocumentNode type of Document
@dc.dataclass()
class DocumentYamlNode(BaseYamlNode):
    start: ta.Optional[YamlToken] = dc.field(default_factory=dataclass_field_required('start'))  # position of DocumentHeader ( `---` )  # noqa
    end: ta.Optional[YamlToken] = None  # position of DocumentEnd ( `...` )
    body: ta.Optional[YamlNode] = dc.field(default_factory=dataclass_field_required('body'))

    # read implements (io.Reader).Read
    def read(self, p: str) -> YamlErrorOr[int]:
        return yaml_read_node(p, self)

    # type returns DocumentNodeType
    def type(self) -> YamlNodeType:
        return YamlNodeType.DOCUMENT

    # get_token returns token instance
    def get_token(self) -> ta.Optional[YamlToken]:
        return check.not_none(self.body).get_token()

    # add_column add column number to child nodes recursively
    def add_column(self, col: int) -> None:
        if self.body is not None:
            self.body.add_column(col)

    # string document to text
    def string(self) -> str:
        doc: ta.List[str] = []
        if self.start is not None:
            doc.append(self.start.value)
        if self.body is not None:
            doc.append(self.body.string())
        if self.end is not None:
            doc.append(self.end.value)
        return '\n'.join(doc)

    # marshal_yaml encodes to a YAML text
    def marshal_yaml(self) -> YamlErrorOr[str]:
        return self.string()


##


# NullNode type of null node
@dc.dataclass()
class NullYamlNode(ScalarYamlNode, BaseYamlNode):
    token: YamlToken = dc.field(default_factory=dataclass_field_required('token'))

    # read implements(io.Reader).Read
    def read(self, p: str) -> YamlErrorOr[int]:
        return yaml_read_node(p, self)

    # type returns NullType
    def type(self) -> YamlNodeType:
        return YamlNodeType.NULL

    # get_token returns token instance
    def get_token(self) -> YamlToken:
        return self.token

    # add_column add column number to child nodes recursively
    def add_column(self, col: int) -> None:
        YamlToken.add_column(self.token, col)

    # get_value returns nil value
    def get_value(self) -> ta.Any:
        return None

    # String returns `null` text
    def string(self) -> str:
        if self.token.type == YamlTokenType.IMPLICIT_NULL:
            if self.comment is not None:
                return self.comment.string()
            return ''
        if self.comment is not None:
            return yaml_add_comment_string('null', self.comment)
        return self.string_without_comment()

    def string_without_comment(self) -> str:
        return 'null'

    # marshal_yaml encodes to a YAML text
    def marshal_yaml(self) -> YamlErrorOr[str]:
        return self.string()

    # is_merge_key returns whether it is a MergeKey node.
    def is_merge_key(self) -> bool:
        return False


##


# IntegerNode type of integer node
@dc.dataclass()
class IntegerYamlNode(ScalarYamlNode, BaseYamlNode):
    token: YamlToken = dc.field(default_factory=dataclass_field_required('token'))
    value: ta.Any = dc.field(default_factory=dataclass_field_required('value'))  # int64 or uint64 value

    # read implements(io.Reader).Read
    def read(self, p: str) -> YamlErrorOr[int]:
        return yaml_read_node(p, self)

    # type returns IntegerType
    def type(self) -> YamlNodeType:
        return YamlNodeType.INTEGER

    # get_token returns token instance
    def get_token(self) -> YamlToken:
        return self.token

    # add_column add column number to child nodes recursively
    def add_column(self, col: int) -> None:
        YamlToken.add_column(self.token, col)

    # get_value returns int64 value
    def get_value(self) -> ta.Any:
        return self.value

    # String int64 to text
    def string(self) -> str:
        if self.comment is not None:
            return yaml_add_comment_string(self.token.value, self.comment)
        return self.string_without_comment()

    def string_without_comment(self) -> str:
        return self.token.value

    # marshal_yaml encodes to a YAML text
    def marshal_yaml(self) -> YamlErrorOr[str]:
        return self.string()

    # is_merge_key returns whether it is a MergeKey node.
    def is_merge_key(self) -> bool:
        return False


##


# FloatNode type of float node
@dc.dataclass()
class FloatYamlNode(ScalarYamlNode, BaseYamlNode):
    token: YamlToken = dc.field(default_factory=dataclass_field_required('token'))
    precision: int = 0
    value: float = dc.field(default_factory=dataclass_field_required('value'))

    # read implements(io.Reader).Read
    def read(self, p: str) -> YamlErrorOr[int]:
        return yaml_read_node(p, self)

    # type returns FloatType
    def type(self) -> YamlNodeType:
        return YamlNodeType.FLOAT

    # get_token returns token instance
    def get_token(self) -> YamlToken:
        return self.token

    # add_column add column number to child nodes recursively
    def add_column(self, col: int) -> None:
        YamlToken.add_column(self.token, col)

    # get_value returns float64 value
    def get_value(self) -> ta.Any:
        return self.value

    # String float64 to text
    def string(self) -> str:
        if self.comment is not None:
            return yaml_add_comment_string(self.token.value, self.comment)
        return self.string_without_comment()

    def string_without_comment(self) -> str:
        return self.token.value

    # marshal_yaml encodes to a YAML text
    def marshal_yaml(self) -> YamlErrorOr[str]:
        return self.string()

    # is_merge_key returns whether it is a MergeKey node.
    def is_merge_key(self) -> bool:
        return False


##


def _yaml_go_is_print(char_ord):
    """
    Approximates Go's unicode.IsPrint logic. A rune is printable if it is a letter, mark, number, punctuation, symbol,
    or ASCII space. (Corresponds to Unicode categories L, M, N, P, S, plus U+0020 SPACE).
    """

    if char_ord == 0x20:  # ASCII space
        return True
    # Check if the character is in categories L, M, N, P, S (Graphic characters)
    category = unicodedata.category(chr(char_ord))
    if category.startswith(('L', 'M', 'N', 'P', 'S')):
        return True
    return False


def _yaml_strconv_quote(s: str) -> str:
    """
    Produces a double-quoted string literal with Go-style escapes, similar to Go's strconv.Quote.
    """

    res = ['"']
    for char_val in s:
        char_ord = ord(char_val)

        if char_val == '"':
            res.append('\\"')
        elif char_val == '\\':
            res.append('\\\\')
        elif char_val == '\a':
            res.append('\\a')
        elif char_val == '\b':
            res.append('\\b')
        elif char_val == '\f':
            res.append('\\f')
        elif char_val == '\n':
            res.append('\\n')
        elif char_val == '\r':
            res.append('\\r')
        elif char_val == '\t':
            res.append('\\t')
        elif char_val == '\v':
            res.append('\\v')
        elif char_ord < 0x20 or char_ord == 0x7F:  # C0 controls and DEL
            res.append(f'\\x{char_ord:02x}')
        elif 0x20 <= char_ord < 0x7F:  # Printable ASCII (already handled \, ")
            res.append(char_val)
        # Unicode characters (char_ord >= 0x80) and C1 controls (0x80-0x9F)
        elif _yaml_go_is_print(char_ord):
            res.append(char_val)
        elif char_ord <= 0xFFFF:
            res.append(f'\\u{char_ord:04x}')
        else:
            res.append(f'\\U{char_ord:08x}')

    res.append('"')
    return ''.join(res)


##


# StringNode type of string node
@dc.dataclass()
class StringYamlNode(ScalarYamlNode, BaseYamlNode):
    token: YamlToken = dc.field(default_factory=dataclass_field_required('token'))
    value: str = dc.field(default_factory=dataclass_field_required('value'))

    # read implements(io.Reader).Read
    def read(self, p: str) -> YamlErrorOr[int]:
        return yaml_read_node(p, self)

    # type returns StringType
    def type(self) -> YamlNodeType:
        return YamlNodeType.STRING

    # get_token returns token instance
    def get_token(self) -> YamlToken:
        return self.token

    # add_column add column number to child nodes recursively
    def add_column(self, col: int) -> None:
        YamlToken.add_column(self.token, col)

    # get_value returns string value
    def get_value(self) -> ta.Any:
        return self.value

    # is_merge_key returns whether it is a MergeKey node.
    def is_merge_key(self) -> bool:
        return False

    # string string value to text with quote or literal header if required
    def string(self) -> str:
        if self.token.type == YamlTokenType.SINGLE_QUOTE:
            quoted = _yaml_escape_single_quote(self.value)
            if self.comment is not None:
                return yaml_add_comment_string(quoted, self.comment)
            return quoted
        elif self.token.type == YamlTokenType.DOUBLE_QUOTE:
            quoted = _yaml_strconv_quote(self.value)
            if self.comment is not None:
                return yaml_add_comment_string(quoted, self.comment)
            return quoted

        lbc = yaml_detect_line_break_char(self.value)
        if lbc in self.value:
            # This block assumes that the line breaks in this inside scalar content and the Outside scalar content are
            # the same. It works mostly, but inconsistencies occur if line break characters are mixed.
            header = yaml_literal_block_header(self.value)
            space = ' ' * (self.token.position.column - 1)
            indent = ' ' * self.token.position.indent_num
            values: ta.List[str] = []
            for v in self.value.split(lbc):
                values.append(f'{space}{indent}{v}')
            block = lbc.join(values).rstrip(f'{lbc}{indent}{space}').rstrip(f'{indent}{space}')
            return f'{header}{lbc}{block}'
        elif len(self.value) > 0 and (self.value[0] == '{' or self.value[0] == '['):
            return f"'{self.value}'"
        if self.comment is not None:
            return yaml_add_comment_string(self.value, self.comment)
        return self.value

    def string_without_comment(self) -> str:
        if self.token.type == YamlTokenType.SINGLE_QUOTE:
            quoted = f"'{self.value}'"
            return quoted
        elif self.token.type == YamlTokenType.DOUBLE_QUOTE:
            quoted = _yaml_strconv_quote(self.value)
            return quoted

        lbc = yaml_detect_line_break_char(self.value)
        if lbc in self.value:
            # This block assumes that the line breaks in this inside scalar content and the Outside scalar content are
            # the same. It works mostly, but inconsistencies occur if line break characters are mixed.
            header = yaml_literal_block_header(self.value)
            space = ' ' * (self.token.position.column - 1)
            indent = ' ' * self.token.position.indent_num
            values: ta.List[str] = []
            for v in self.value.split(lbc):
                values.append(f'{space}{indent}{v}')
            block = lbc.join(values).rstrip(f'{lbc}{indent}{space}').rstrip(f'  {space}')
            return f'{header}{lbc}{block}'
        elif len(self.value) > 0 and (self.value[0] == '{' or self.value[0] == '['):
            return f"'{self.value}'"
        return self.value

    # marshal_yaml encodes to a YAML text
    def marshal_yaml(self) -> YamlErrorOr[str]:
        return self.string()


# escape_single_quote escapes s to a single quoted scalar.
# https://yaml.org/spec/1.2.2/#732-single-quoted-style
def _yaml_escape_single_quote(s: str) -> str:
    sb = io.StringIO()
    # growLen = len(s) + # s includes also one ' from the doubled pair
    #     2 + # opening and closing '
    #     strings.Count(s, "'") # ' added by ReplaceAll
    # sb.Grow(growLen)
    sb.write("'")
    sb.write(s.replace("'", "''"))
    sb.write("'")
    return sb.getvalue()


##


# LiteralNode type of literal node
@dc.dataclass()
class LiteralYamlNode(ScalarYamlNode, BaseYamlNode):
    start: YamlToken = dc.field(default_factory=dataclass_field_required('start'))
    value: ta.Optional['StringYamlNode'] = None

    # read implements(io.Reader).Read
    def read(self, p: str) -> YamlErrorOr[int]:
        return yaml_read_node(p, self)

    # type returns LiteralType
    def type(self) -> YamlNodeType:
        return YamlNodeType.LITERAL

    # get_token returns token instance
    def get_token(self) -> YamlToken:
        return self.start

    # add_column add column number to child nodes recursively
    def add_column(self, col: int) -> None:
        YamlToken.add_column(self.start, col)
        if self.value is not None:
            self.value.add_column(col)

    # get_value returns string value
    def get_value(self) -> ta.Any:
        return self.string()

    # String literal to text
    def string(self) -> str:
        origin = check.not_none(check.not_none(self.value).get_token()).origin
        lit = origin.rstrip(' ').rstrip('\n')
        if self.comment is not None:
            return f'{self.start.value} {self.comment.string()}\n{lit}'
        return f'{self.start.value}\n{lit}'

    def string_without_comment(self) -> str:
        return self.string()

    # marshal_yaml encodes to a YAML text
    def marshal_yaml(self) -> YamlErrorOr[str]:
        return self.string()

    # is_merge_key returns whether it is a MergeKey node.
    def is_merge_key(self) -> bool:
        return False


##


# MergeKeyNode type of merge key node
@dc.dataclass()
class MergeKeyYamlNode(ScalarYamlNode, BaseYamlNode):
    token: YamlToken = dc.field(default_factory=dataclass_field_required('token'))

    # read implements(io.Reader).Read
    def read(self, p: str) -> YamlErrorOr[int]:
        return yaml_read_node(p, self)

    # type returns MergeKeyType
    def type(self) -> YamlNodeType:
        return YamlNodeType.MERGE_KEY

    # get_token returns token instance
    def get_token(self) -> YamlToken:
        return self.token

    # get_value returns '<<' value
    def get_value(self) -> ta.Any:
        return self.token.value

    # String returns '<<' value
    def string(self) -> str:
        return self.string_without_comment()

    def string_without_comment(self) -> str:
        return self.token.value

    # add_column add column number to child nodes recursively
    def add_column(self, col: int) -> None:
        YamlToken.add_column(self.token, col)

    # marshal_yaml encodes to a YAML text
    def marshal_yaml(self) -> YamlErrorOr[str]:
        return str(self)

    # is_merge_key returns whether it is a MergeKey node.
    def is_merge_key(self) -> bool:
        return True


##


# BoolNode type of boolean node
@dc.dataclass()
class BoolYamlNode(ScalarYamlNode, BaseYamlNode):
    token: YamlToken = dc.field(default_factory=dataclass_field_required('token'))
    value: bool = dc.field(default_factory=dataclass_field_required('value'))

    # read implements(io.Reader).Read
    def read(self, p: str) -> YamlErrorOr[int]:
        return yaml_read_node(p, self)

    # type returns BoolType
    def type(self) -> YamlNodeType:
        return YamlNodeType.BOOL

    # get_token returns token instance
    def get_token(self) -> YamlToken:
        return self.token

    # add_column add column number to child nodes recursively
    def add_column(self, col: int) -> None:
        YamlToken.add_column(self.token, col)

    # get_value returns boolean value
    def get_value(self) -> ta.Any:
        return self.value

    # String boolean to text
    def string(self) -> str:
        if self.comment is not None:
            return yaml_add_comment_string(self.token.value, self.comment)
        return self.string_without_comment()

    def string_without_comment(self) -> str:
        return self.token.value

    # marshal_yaml encodes to a YAML text
    def marshal_yaml(self) -> YamlErrorOr[str]:
        return self.string()

    # is_merge_key returns whether it is a MergeKey node.
    def is_merge_key(self) -> bool:
        return False


##


# InfinityNode type of infinity node
@dc.dataclass()
class InfinityYamlNode(ScalarYamlNode, BaseYamlNode):
    token: YamlToken = dc.field(default_factory=dataclass_field_required('token'))
    value: float = dc.field(default_factory=dataclass_field_required('value'))

    # read implements(io.Reader).Read
    def read(self, p: str) -> YamlErrorOr[int]:
        return yaml_read_node(p, self)

    # type returns InfinityType
    def type(self) -> YamlNodeType:
        return YamlNodeType.INFINITY

    # get_token returns token instance
    def get_token(self) -> YamlToken:
        return self.token

    # add_column add column number to child nodes recursively
    def add_column(self, col: int) -> None:
        YamlToken.add_column(self.token, col)

    # get_value returns math.Inf(0) or math.Inf(-1)
    def get_value(self) -> ta.Any:
        return self.value

    # String infinity to text
    def string(self) -> str:
        if self.comment is not None:
            return yaml_add_comment_string(self.token.value, self.comment)
        return self.string_without_comment()

    def string_without_comment(self) -> str:
        return self.token.value

    # marshal_yaml encodes to a YAML text
    def marshal_yaml(self) -> YamlErrorOr[str]:
        return self.string()

    # is_merge_key returns whether it is a MergeKey node.
    def is_merge_key(self) -> bool:
        return False


##


# NanNode type of nan node
@dc.dataclass()
class NanYamlNode(ScalarYamlNode, BaseYamlNode):
    token: YamlToken = dc.field(default_factory=dataclass_field_required('token'))

    # read implements(io.Reader).Read
    def read(self, p: str) -> YamlErrorOr[int]:
        return yaml_read_node(p, self)

    # type returns NanType
    def type(self) -> YamlNodeType:
        return YamlNodeType.NAN

    # get_token returns token instance
    def get_token(self) -> YamlToken:
        return self.token

    # add_column add column number to child nodes recursively
    def add_column(self, col: int) -> None:
        YamlToken.add_column(self.token, col)

    # get_value returns math.NaN()
    def get_value(self) -> ta.Any:
        return float('nan')

    # String returns .nan
    def string(self) -> str:
        if self.comment is not None:
            return yaml_add_comment_string(self.token.value, self.comment)
        return self.string_without_comment()

    def string_without_comment(self) -> str:
        return self.token.value

    # marshal_yaml encodes to a YAML text
    def marshal_yaml(self) -> YamlErrorOr[str]:
        return self.string()

    # is_merge_key returns whether it is a MergeKey node.
    def is_merge_key(self) -> bool:
        return False


##


# MapNode interface of MappingValueNode / MappingNode
class MapYamlNode(Abstract):
    @abc.abstractmethod
    def map_range(self) -> 'MapYamlNodeIter':
        raise NotImplementedError


YAML_START_RANGE_INDEX = -1


# MapNodeIter is an iterator for ranging over a MapNode
@dc.dataclass()
class MapYamlNodeIter:
    values: ta.List['MappingValueYamlNode']
    idx: int

    # next advances the map iterator and reports whether there is another entry.
    # It returns false when the iterator is exhausted.
    def next(self) -> bool:
        self.idx += 1
        nxt = self.idx < len(self.values)
        return nxt

    # key returns the key of the iterator's current map node entry.
    def key(self) -> MapKeyYamlNode:
        return self.values[self.idx].key

    # value returns the value of the iterator's current map node entry.
    def value(self) -> YamlNode:
        return self.values[self.idx].value

    # key_value returns the MappingValueNode of the iterator's current map node entry.
    def key_value(self) -> 'MappingValueYamlNode':
        return self.values[self.idx]


#


# MappingNode type of mapping node
@dc.dataclass()
class MappingYamlNode(MapYamlNode, BaseYamlNode):
    start: YamlToken = dc.field(default_factory=dataclass_field_required('start'))
    end: ta.Optional[YamlToken] = None
    is_flow_style: bool = dc.field(default_factory=dataclass_field_required('is_flow_style'))
    values: ta.List['MappingValueYamlNode'] = dc.field(default_factory=dataclass_field_required('values'))
    foot_comment: ta.Optional['CommentGroupYamlNode'] = None

    def start_pos(self) -> YamlPosition:
        if len(self.values) == 0:
            return self.start.position
        return check.not_none(self.values[0].key.get_token()).position

    # merge merge key/value of map.
    def merge(self, target: 'MappingYamlNode') -> None:
        key_to_map_value_map: ta.Dict[str, MappingValueYamlNode] = {}
        for value in self.values:
            key = value.key.string()
            key_to_map_value_map[key] = value
        column = self.start_pos().column - target.start_pos().column
        target.add_column(column)
        for value in target.values:
            map_value = key_to_map_value_map.get(value.key.string())
            if map_value is not None:
                map_value.value = value.value
            else:
                self.values.append(value)

    # set_is_flow_style set value to is_flow_style field recursively.
    def set_is_flow_style(self, is_flow: bool) -> None:
        self.is_flow_style = is_flow
        for value in self.values:
            value.set_is_flow_style(is_flow)

    # read implements(io.Reader).Read
    def read(self, p: str) -> YamlErrorOr[int]:
        return yaml_read_node(p, self)

    # type returns MappingType
    def type(self) -> YamlNodeType:
        return YamlNodeType.MAPPING

    # get_token returns token instance
    def get_token(self) -> YamlToken:
        return self.start

    # add_column add column number to child nodes recursively
    def add_column(self, col: int) -> None:
        YamlToken.add_column(self.start, col)
        YamlToken.add_column(self.end, col)
        for value in self.values:
            value.add_column(col)

    def flow_style_string(self, comment_mode: bool) -> str:
        values: ta.List[str] = []
        for value in self.values:
            values.append(value.string().lstrip(' '))
        map_text = f'{{{", ".join(values)}}}'
        if comment_mode and self.comment is not None:
            return yaml_add_comment_string(map_text, self.comment)
        return map_text

    def block_style_string(self, comment_mode: bool) -> str:
        values: ta.List[str] = []
        for value0 in self.values:
            values.append(value0.string())
        map_text = '\n'.join(values)
        if comment_mode and self.comment is not None:
            value1 = values[0]
            space_num = 0
            for i in range(len(value1)):
                if value1[i] != ' ':
                    break
                space_num += 1
            comment = self.comment.string_with_space(space_num)
            return f'{comment}\n{map_text}'
        return map_text

    # String mapping values to text
    def string(self) -> str:
        if len(self.values) == 0:
            if self.comment is not None:
                return yaml_add_comment_string('{}', self.comment)
            return '{}'

        comment_mode = True
        if self.is_flow_style or len(self.values) == 0:
            return self.flow_style_string(comment_mode)

        return self.block_style_string(comment_mode)

    # map_range implements MapNode protocol
    def map_range(self) -> MapYamlNodeIter:
        return MapYamlNodeIter(
            idx=YAML_START_RANGE_INDEX,
            values=self.values,
        )

    # marshal_yaml encodes to a YAML text
    def marshal_yaml(self) -> YamlErrorOr[str]:
        return self.string()


##


# MappingKeyNode type of tag node
@dc.dataclass()
class MappingKeyYamlNode(MapKeyYamlNode, BaseYamlNode):
    start: YamlToken = dc.field(default_factory=dataclass_field_required('start'))
    value: ta.Optional[YamlNode] = None

    # read implements(io.Reader).Read
    def read(self, p: str) -> YamlErrorOr[int]:
        return yaml_read_node(p, self)

    # type returns MappingKeyType
    def type(self) -> YamlNodeType:
        return YamlNodeType.MAPPING_KEY

    # get_token returns token instance
    def get_token(self) -> YamlToken:
        return self.start

    # add_column add column number to child nodes recursively
    def add_column(self, col: int) -> None:
        YamlToken.add_column(self.start, col)
        if self.value is not None:
            self.value.add_column(col)

    # String tag to text
    def string(self) -> str:
        return self.string_without_comment()

    def string_without_comment(self) -> str:
        return f'{self.start.value} {check.not_none(self.value).string()}'

    # marshal_yaml encodes to a YAML text
    def marshal_yaml(self) -> YamlErrorOr[str]:
        return self.string()

    # is_merge_key returns whether it is a MergeKey node.
    def is_merge_key(self) -> bool:
        if self.value is None:
            return False
        key = self.value
        if not isinstance(key, MapKeyYamlNode):
            return False
        return key.is_merge_key()


##


# MappingValueNode type of mapping value
@dc.dataclass()
class MappingValueYamlNode(MapYamlNode, BaseYamlNode):
    start: YamlToken = dc.field(default_factory=dataclass_field_required('start'))  # delimiter token ':'.
    collect_entry: ta.Optional[YamlToken] = None  # collect entry token ','.
    key: MapKeyYamlNode = dc.field(default_factory=dataclass_field_required('key'))
    value: YamlNode = dc.field(default_factory=dataclass_field_required('value'))
    foot_comment: ta.Optional['CommentGroupYamlNode'] = None
    is_flow_style: bool = False

    # Replace replace value node.
    def replace(self, value: YamlNode) -> ta.Optional[YamlError]:
        column = check.not_none(self.value.get_token()).position.column - check.not_none(value.get_token()).position.column  # noqa
        value.add_column(column)
        self.value = value
        return None

    # read implements(io.Reader).Read
    def read(self, p: str) -> YamlErrorOr[int]:
        return yaml_read_node(p, self)

    # type returns MappingValueType
    def type(self) -> YamlNodeType:
        return YamlNodeType.MAPPING_VALUE

    # get_token returns token instance
    def get_token(self) -> YamlToken:
        return self.start

    # add_column add column number to child nodes recursively
    def add_column(self, col: int) -> None:
        YamlToken.add_column(self.start, col)
        if self.key is not None:
            self.key.add_column(col)
        if self.value is not None:
            self.value.add_column(col)

    # set_is_flow_style set value to is_flow_style field recursively.
    def set_is_flow_style(self, is_flow: bool) -> None:
        self.is_flow_style = is_flow
        if isinstance(self.value, MappingYamlNode):
            self.value.set_is_flow_style(is_flow)
        elif isinstance(self.value, MappingValueYamlNode):
            self.value.set_is_flow_style(is_flow)
        elif isinstance(self.value, SequenceYamlNode):
            self.value.set_is_flow_style(is_flow)

    # String mapping value to text
    def string(self) -> str:
        text: str
        if self.comment is not None:
            text = f'{self.comment.string_with_space(check.not_none(self.key.get_token()).position.column - 1)}\n{self.to_string()}'  # noqa
        else:
            text = self.to_string()

        if self.foot_comment is not None:
            text += f'\n{self.foot_comment.string_with_space(check.not_none(self.key.get_token()).position.column - 1)}'

        return text

    def to_string(self) -> str:
        space = ' ' * (check.not_none(self.key.get_token()).position.column - 1)
        if yaml_check_line_break(check.not_none(self.key.get_token())):
            space = f'\n{space}'

        key_indent_level = check.not_none(self.key.get_token()).position.indent_level
        value_indent_level = check.not_none(self.value.get_token()).position.indent_level
        key_comment = self.key.get_comment()

        if isinstance(self.value, ScalarYamlNode):
            value = self.value.string()
            if value == '':
                # implicit null value.
                return f'{space}{self.key.string()}:'
            return f'{space}{self.key.string()}: {value}'

        elif key_indent_level < value_indent_level and not self.is_flow_style:
            value_str = self.value.string()
            # For flow-style values indented on the next line, we need to add the proper indentation
            if isinstance(self.value, MappingYamlNode) and self.value.is_flow_style:
                value_indent = ' ' * (self.value.get_token().position.column - 1)
                value_str = value_indent + value_str
            elif isinstance(self.value, SequenceYamlNode) and self.value.is_flow_style:
                value_indent = ' ' * (self.value.get_token().position.column - 1)
                value_str = value_indent + value_str
            if key_comment is not None:
                return f'{space}{self.key.string_without_comment()}: {key_comment.string()}\n{value_str}'

            return f'{space}{self.key.string()}:\n{value_str}'

        elif isinstance(self.value, MappingYamlNode) and (self.value.is_flow_style or len(self.value.values) == 0):
            return f'{space}{self.key.string()}: {self.value.string()}'

        elif isinstance(self.value, SequenceYamlNode) and (self.value.is_flow_style or len(self.value.values) == 0):
            return f'{space}{self.key.string()}: {self.value.string()}'

        elif isinstance(self.value, AnchorYamlNode):
            return f'{space}{self.key.string()}: {self.value.string()}'

        elif isinstance(self.value, AliasYamlNode):
            return f'{space}{self.key.string()}: {self.value.string()}'

        elif isinstance(self.value, TagYamlNode):
            return f'{space}{self.key.string()}: {self.value.string()}'

        if key_comment is not None:
            return f'{space}{self.key.string_without_comment()}: {key_comment.string()}\n{self.value.string()}'

        if isinstance(self.value, MappingYamlNode) and self.value.comment is not None:
            return f'{space}{self.key.string()}: {self.value.string().lstrip(" ")}'

        return f'{space}{self.key.string()}:\n{self.value.string()}'

    # map_range implements MapNode protocol
    def map_range(self) -> MapYamlNodeIter:
        return MapYamlNodeIter(
            idx=YAML_START_RANGE_INDEX,
            values=[self],
        )

    # marshal_yaml encodes to a YAML text
    def marshal_yaml(self) -> YamlErrorOr[str]:
        return self.string()


##


# ArrayNode interface of SequenceNode
class ArrayYamlNode(YamlNode, Abstract):
    @abc.abstractmethod
    def array_range(self) -> ta.Optional['ArrayYamlNodeIter']:
        raise NotImplementedError


# ArrayNodeIter is an iterator for ranging over a ArrayNode
@dc.dataclass()
class ArrayYamlNodeIter:
    values: ta.List[YamlNode]
    idx: int

    # next advances the array iterator and reports whether there is another entry.
    # It returns false when the iterator is exhausted.
    def next(self) -> bool:
        self.idx += 1
        nxt = self.idx < len(self.values)
        return nxt

    # Value returns the value of the iterator's current array entry.
    def value(self) -> YamlNode:
        return self.values[self.idx]

    # len returns length of array
    def len(self) -> int:
        return len(self.values)


##


# SequenceNode type of sequence node
@dc.dataclass()
class SequenceYamlNode(BaseYamlNode, ArrayYamlNode):
    start: YamlToken = dc.field(default_factory=dataclass_field_required('start'))
    end: ta.Optional[YamlToken] = None
    is_flow_style: bool = dc.field(default_factory=dataclass_field_required('is_flow_style'))
    values: ta.List[ta.Optional[YamlNode]] = dc.field(default_factory=dataclass_field_required('values'))
    value_head_comments: ta.List[ta.Optional['CommentGroupYamlNode']] = dc.field(default_factory=list)
    entries: ta.List['SequenceEntryYamlNode'] = dc.field(default_factory=list)
    foot_comment: ta.Optional['CommentGroupYamlNode'] = None

    # replace replace value node.
    def replace(self, idx: int, value: YamlNode) -> ta.Optional[YamlError]:
        if len(self.values) <= idx:
            return yaml_error(f'invalid index for sequence: sequence length is {len(self.values):d}, but specified {idx:d} index')  # noqa

        column = check.not_none(check.not_none(self.values[idx]).get_token()).position.column - check.not_none(value.get_token()).position.column  # noqa
        value.add_column(column)
        self.values[idx] = value
        return None

    # merge merge sequence value.
    def merge(self, target: 'SequenceYamlNode') -> None:
        column = self.start.position.column - target.start.position.column
        target.add_column(column)
        self.values.extend(target.values)
        if len(target.value_head_comments) == 0:
            self.value_head_comments.extend([None] * len(target.values))
            return

        self.value_head_comments.extend(target.value_head_comments)

    # set_is_flow_style set value to is_flow_style field recursively.
    def set_is_flow_style(self, is_flow: bool) -> None:
        self.is_flow_style = is_flow
        for value in self.values:
            if isinstance(value, MappingYamlNode):
                value.set_is_flow_style(is_flow)
            elif isinstance(value, MappingValueYamlNode):
                value.set_is_flow_style(is_flow)
            elif isinstance(value, SequenceYamlNode):
                value.set_is_flow_style(is_flow)

    # read implements(io.Reader).Read
    def read(self, p: str) -> YamlErrorOr[int]:
        return yaml_read_node(p, self)

    # type returns SequenceType
    def type(self) -> YamlNodeType:
        return YamlNodeType.SEQUENCE

    # get_token returns token instance
    def get_token(self) -> YamlToken:
        return self.start

    # add_column add column number to child nodes recursively
    def add_column(self, col: int) -> None:
        YamlToken.add_column(self.start, col)
        YamlToken.add_column(self.end, col)
        for value in self.values:
            check.not_none(value).add_column(col)

    def flow_style_string(self) -> str:
        values: ta.List[str] = []
        for value in self.values:
            values.append(check.not_none(value).string())

        seq_text = f'[{", ".join(values)}]'
        if self.comment is not None:
            return yaml_add_comment_string(seq_text, self.comment)

        return seq_text

    def block_style_string(self) -> str:
        space = ' ' * (self.start.position.column - 1)
        values: ta.List[str] = []
        if self.comment is not None:
            values.append(self.comment.string_with_space(self.start.position.column - 1))

        for idx, value in enumerate(self.values):
            if value is None:
                continue

            value_str = value.string()
            new_line_prefix = ''
            if value_str.startswith('\n'):
                value_str = value_str[1:]
                new_line_prefix = '\n'

            splitted_values = value_str.split('\n')
            trimmed_first_value = splitted_values[0].lstrip(' ')
            diff_length = len(splitted_values[0]) - len(trimmed_first_value)
            if (
                    (len(splitted_values) > 1 and value.type() == YamlNodeType.STRING) or
                    value.type() == YamlNodeType.LITERAL
            ):
                # If multi-line string, the space characters for indent have already been added, so delete them.
                prefix = space + '  '
                for i in range(1, len(splitted_values)):
                    splitted_values[i] = splitted_values[i].lstrip(prefix)

            new_values: ta.List[str] = [trimmed_first_value]
            for i in range(1, len(splitted_values)):
                if len(splitted_values[i]) <= diff_length:
                    # this line is \n or white space only
                    new_values.append('')
                    continue

                trimmed = splitted_values[i][diff_length:]
                new_values.append(f'{space}  {trimmed}')

            new_value = '\n'.join(new_values)
            if len(self.value_head_comments) == len(self.values) and self.value_head_comments[idx] is not None:
                values.append(
                    f'{new_line_prefix}'
                    f'{check.not_none(self.value_head_comments[idx]).string_with_space(self.start.position.column - 1)}',  # noqa
                )
                new_line_prefix = ''

            values.append(f'{new_line_prefix}{space}- {new_value}')

        if self.foot_comment is not None:
            values.append(self.foot_comment.string_with_space(self.start.position.column - 1))

        return '\n'.join(values)

    # String sequence to text
    def string(self) -> str:
        if self.is_flow_style or len(self.values) == 0:
            return self.flow_style_string()
        return self.block_style_string()

    # array_range implements ArrayNode protocol
    def array_range(self) -> ta.Optional[ArrayYamlNodeIter]:
        return ArrayYamlNodeIter(
            idx=YAML_START_RANGE_INDEX,
            values=ta.cast('ta.List[YamlNode]', self.values),
        )

    # marshal_yaml encodes to a YAML text
    def marshal_yaml(self) -> YamlErrorOr[str]:
        return self.string()


##


# SequenceEntryNode is the sequence entry.
@dc.dataclass()
class SequenceEntryYamlNode(BaseYamlNode):
    head_comment: ta.Optional['CommentGroupYamlNode'] = dc.field(default_factory=dataclass_field_required('head_commend'))  # head comment.  # noqa
    line_comment: ta.Optional['CommentGroupYamlNode'] = None  # line comment e.g.) - # comment.
    start: ta.Optional[YamlToken] = dc.field(default_factory=dataclass_field_required('start'))  # entry token.
    value: YamlNode = dc.field(default_factory=dataclass_field_required('value'))  # value node.

    # String node to text
    def string(self) -> str:
        return ''  # TODO

    # get_token returns token instance
    def get_token(self) -> ta.Optional[YamlToken]:
        return self.start

    # type returns type of node
    def type(self) -> YamlNodeType:
        return YamlNodeType.SEQUENCE_ENTRY

    # add_column add column number to child nodes recursively
    def add_column(self, col: int) -> None:
        YamlToken.add_column(self.start, col)

    # set_comment set line comment.
    def set_comment(self, node: ta.Optional['CommentGroupYamlNode']) -> ta.Optional[YamlError]:
        self.line_comment = node
        return None

    # comment returns comment token instance
    def get_comment(self) -> ta.Optional['CommentGroupYamlNode']:
        return self.line_comment

    # marshal_yaml
    def marshal_yaml(self) -> YamlErrorOr[str]:
        return self.string()

    def read(self, p: str) -> YamlErrorOr[int]:
        return yaml_read_node(p, self)


# sequence_entry creates SequenceEntryNode instance.
def yaml_sequence_entry(
        start: ta.Optional[YamlToken],
        value: YamlNode,
        head_comment: ta.Optional['CommentGroupYamlNode'],
) -> SequenceEntryYamlNode:
    return SequenceEntryYamlNode(
        head_comment=head_comment,
        start=start,
        value=value,
    )


# SequenceMergeValue creates SequenceMergeValueNode instance.
def yaml_sequence_merge_value(*values: MapYamlNode) -> 'SequenceMergeValueYamlNode':
    return SequenceMergeValueYamlNode(
        values=list(values),
    )


##


# SequenceMergeValueNode is used to convert the Sequence node specified for the merge key into a MapNode format.
@dc.dataclass()
class SequenceMergeValueYamlNode(MapYamlNode):
    values: ta.List[MapYamlNode] = dc.field(default_factory=dataclass_field_required('values'))

    # map_range returns MapNodeIter instance.
    def map_range(self) -> MapYamlNodeIter:
        ret = MapYamlNodeIter(values=[], idx=YAML_START_RANGE_INDEX)
        for value in self.values:
            it = value.map_range()
            ret.values.extend(it.values)
        return ret


##


# AnchorNode type of anchor node
@dc.dataclass()
class AnchorYamlNode(ScalarYamlNode, BaseYamlNode):
    start: YamlToken = dc.field(default_factory=dataclass_field_required('start'))
    name: ta.Optional[YamlNode] = None
    value: ta.Optional[YamlNode] = None

    def string_without_comment(self) -> str:
        return check.not_none(self.value).string()

    def set_name(self, name: str) -> ta.Optional[YamlError]:
        if self.name is None:
            return YamlAstErrors.INVALID_ANCHOR_NAME
        s = self.name
        if not isinstance(s, StringYamlNode):
            return YamlAstErrors.INVALID_ANCHOR_NAME
        s.value = name
        return None

    # read implements(io.Reader).Read
    def read(self, p: str) -> YamlErrorOr[int]:
        return yaml_read_node(p, self)

    # type returns AnchorType
    def type(self) -> YamlNodeType:
        return YamlNodeType.ANCHOR

    # get_token returns token instance
    def get_token(self) -> YamlToken:
        return self.start

    def get_value(self) -> ta.Any:
        return check.not_none(check.not_none(self.value).get_token()).value

    # add_column add column number to child nodes recursively
    def add_column(self, col: int) -> None:
        YamlToken.add_column(self.start, col)
        if self.name is not None:
            self.name.add_column(col)
        if self.value is not None:
            self.value.add_column(col)

    # String anchor to text
    def string(self) -> str:
        anchor = '&' + check.not_none(self.name).string()
        value = check.not_none(self.value).string()
        if isinstance(self.value, SequenceYamlNode) and not self.value.is_flow_style:
            return f'{anchor}\n{value}'
        elif isinstance(self.value, MappingYamlNode) and not self.value.is_flow_style:
            return f'{anchor}\n{value}'
        if value == '':
            # implicit null value.
            return anchor
        return f'{anchor} {value}'

    # marshal_yaml encodes to a YAML text
    def marshal_yaml(self) -> YamlErrorOr[str]:
        return self.string()

    # is_merge_key returns whether it is a MergeKey node.
    def is_merge_key(self) -> bool:
        if self.value is None:
            return False
        key = self.value
        if not isinstance(key, MapKeyYamlNode):
            return False
        return key.is_merge_key()


##


# AliasNode type of alias node
@dc.dataclass()
class AliasYamlNode(ScalarYamlNode, BaseYamlNode):
    start: YamlToken = dc.field(default_factory=dataclass_field_required('start'))
    value: ta.Optional[YamlNode] = None

    def string_without_comment(self) -> str:
        return check.not_none(self.value).string()

    def set_name(self, name: str) -> ta.Optional[YamlError]:
        if self.value is None:
            return YamlAstErrors.INVALID_ALIAS_NAME
        if not isinstance(self.value, StringYamlNode):
            return YamlAstErrors.INVALID_ALIAS_NAME
        self.value.value = name
        return None

    # read implements(io.Reader).Read
    def read(self, p: str) -> YamlErrorOr[int]:
        return yaml_read_node(p, self)

    # type returns AliasType
    def type(self) -> YamlNodeType:
        return YamlNodeType.ALIAS

    # get_token returns token instance
    def get_token(self) -> YamlToken:
        return self.start

    def get_value(self) -> ta.Any:
        return check.not_none(check.not_none(self.value).get_token()).value

    # add_column add column number to child nodes recursively
    def add_column(self, col: int) -> None:
        YamlToken.add_column(self.start, col)
        if self.value is not None:
            self.value.add_column(col)

    # String alias to text
    def string(self) -> str:
        return f'*{check.not_none(self.value).string()}'

    # marshal_yaml encodes to a YAML text
    def marshal_yaml(self) -> YamlErrorOr[str]:
        return self.string()

    # is_merge_key returns whether it is a MergeKey node.
    def is_merge_key(self) -> bool:
        return False


##


# DirectiveNode type of directive node
@dc.dataclass()
class DirectiveYamlNode(BaseYamlNode):
    # Start is '%' token.
    start: YamlToken = dc.field(default_factory=dataclass_field_required('start'))
    # Name is directive name e.g.) "YAML" or "TAG".
    name: ta.Optional[YamlNode] = None
    # Values is directive values e.g.) "1.2" or "!!" and "tag:clarkevans.com,2002:app/".
    values: ta.List[YamlNode] = dc.field(default_factory=list)

    # read implements(io.Reader).Read
    def read(self, p: str) -> YamlErrorOr[int]:
        return yaml_read_node(p, self)

    # type returns DirectiveType
    def type(self) -> YamlNodeType:
        return YamlNodeType.DIRECTIVE

    # get_token returns token instance
    def get_token(self) -> YamlToken:
        return self.start

    # add_column add column number to child nodes recursively
    def add_column(self, col: int) -> None:
        if self.name is not None:
            self.name.add_column(col)
        for value in self.values:
            value.add_column(col)

    # String directive to text
    def string(self) -> str:
        values: ta.List[str] = []
        for val in self.values:
            values.append(val.string())
        return ' '.join(['%' + check.not_none(self.name).string(), *values])

    # marshal_yaml encodes to a YAML text
    def marshal_yaml(self) -> YamlErrorOr[str]:
        return self.string()


##


# TagNode type of tag node
@dc.dataclass()
class TagYamlNode(ScalarYamlNode, BaseYamlNode, ArrayYamlNode):
    directive: ta.Optional[DirectiveYamlNode] = None
    start: YamlToken = dc.field(default_factory=dataclass_field_required('start'))
    value: ta.Optional[YamlNode] = None

    def get_value(self) -> ta.Any:
        if not isinstance(self.value, ScalarYamlNode):
            return None
        return self.value.get_value()

    def string_without_comment(self) -> str:
        return check.not_none(self.value).string()

    # read implements(io.Reader).Read
    def read(self, p: str) -> YamlErrorOr[int]:
        return yaml_read_node(p, self)

    # type returns TagType
    def type(self) -> YamlNodeType:
        return YamlNodeType.TAG

    # get_token returns token instance
    def get_token(self) -> YamlToken:
        return self.start

    # add_column add column number to child nodes recursively
    def add_column(self, col: int) -> None:
        YamlToken.add_column(self.start, col)
        if self.value is not None:
            self.value.add_column(col)

    # String tag to text
    def string(self) -> str:
        value = check.not_none(self.value).string()
        if isinstance(self.value, SequenceYamlNode) and not self.value.is_flow_style:
            return f'{self.start.value}\n{value}'
        elif isinstance(self.value, MappingYamlNode) and not self.value.is_flow_style:
            return f'{self.start.value}\n{value}'

        return f'{self.start.value} {value}'

    # marshal_yaml encodes to a YAML text
    def marshal_yaml(self) -> YamlErrorOr[str]:
        return self.string()

    # is_merge_key returns whether it is a MergeKey node.
    def is_merge_key(self) -> bool:
        if self.value is None:
            return False
        key = self.value
        if not isinstance(key, MapKeyYamlNode):
            return False
        return key.is_merge_key()

    def array_range(self) -> ta.Optional[ArrayYamlNodeIter]:
        arr = self.value
        if not isinstance(arr, ArrayYamlNode):
            return None
        return arr.array_range()


##


# CommentNode type of comment node
@dc.dataclass()
class CommentYamlNode(BaseYamlNode):
    token: ta.Optional[YamlToken] = dc.field(default_factory=dataclass_field_required('token'))

    # read implements(io.Reader).Read
    def read(self, p: str) -> YamlErrorOr[int]:
        return yaml_read_node(p, self)

    # type returns CommentType
    def type(self) -> YamlNodeType:
        return YamlNodeType.COMMENT

    # get_token returns token instance
    def get_token(self) -> ta.Optional[YamlToken]:
        return self.token

    # add_column add column number to child nodes recursively
    def add_column(self, col: int) -> None:
        YamlToken.add_column(self.token, col)

    # String comment to text
    def string(self) -> str:
        return f'#{check.not_none(self.token).value}'

    # marshal_yaml encodes to a YAML text
    def marshal_yaml(self) -> YamlErrorOr[str]:
        return self.string()


##


# CommentGroupNode type of comment node
@dc.dataclass()
class CommentGroupYamlNode(BaseYamlNode):
    comments: ta.List[CommentYamlNode] = dc.field(default_factory=dataclass_field_required('comments'))

    # read implements(io.Reader).Read
    def read(self, p: str) -> YamlErrorOr[int]:
        return yaml_read_node(p, self)

    # type returns CommentType
    def type(self) -> YamlNodeType:
        return YamlNodeType.COMMENT

    # get_token returns token instance
    def get_token(self) -> ta.Optional[YamlToken]:
        if len(self.comments) > 0:
            return self.comments[0].token
        return None

    # add_column add column number to child nodes recursively
    def add_column(self, col: int) -> None:
        for comment in self.comments:
            comment.add_column(col)

    # String comment to text
    def string(self) -> str:
        values: ta.List[str] = []
        for comment in self.comments:
            values.append(comment.string())
        return '\n'.join(values)

    def string_with_space(self, col: int) -> str:
        values: ta.List[str] = []
        space = ' ' * col
        for comment in self.comments:
            spc = space
            if yaml_check_line_break(check.not_none(comment.token)):
                spc = f'\n{spc}'
            values.append(spc + comment.string())
        return '\n'.join(values)

    # marshal_yaml encodes to a YAML text
    def marshal_yaml(self) -> YamlErrorOr[str]:
        return self.string()


##


# Visitor has Visit method that is invokded for each node encountered by walk.
# If the result visitor w is not nil, walk visits each of the children of node with the visitor w,
# followed by a call of w.visit(nil).
class YamlAstVisitor(Abstract):
    @abc.abstractmethod
    def visit(self, node: YamlNode) -> ta.Optional['YamlAstVisitor']:
        raise NotImplementedError


# walk traverses an AST in depth-first order: It starts by calling v.visit(node); node must not be nil.
# If the visitor w returned by v.visit(node) is not nil,
# walk is invoked recursively with visitor w for each of the non-nil children of node,
# followed by a call of w.visit(nil).
def yaml_ast_walk(v: YamlAstVisitor, node: YamlNode) -> None:
    if (v_ := v.visit(node)) is None:
        return
    v = v_

    n = node
    if isinstance(n, (CommentYamlNode, NullYamlNode)):
        yaml_ast_walk_comment(v, n)
    if isinstance(n, IntegerYamlNode):
        yaml_ast_walk_comment(v, n)
    if isinstance(n, FloatYamlNode):
        yaml_ast_walk_comment(v, n)
    if isinstance(n, StringYamlNode):
        yaml_ast_walk_comment(v, n)
    if isinstance(n, MergeKeyYamlNode):
        yaml_ast_walk_comment(v, n)
    if isinstance(n, BoolYamlNode):
        yaml_ast_walk_comment(v, n)
    if isinstance(n, InfinityYamlNode):
        yaml_ast_walk_comment(v, n)
    if isinstance(n, NanYamlNode):
        yaml_ast_walk_comment(v, n)
    if isinstance(n, LiteralYamlNode):
        yaml_ast_walk_comment(v, n)
        yaml_ast_walk(v, check.not_none(n.value))
    if isinstance(n, DirectiveYamlNode):
        yaml_ast_walk_comment(v, n)
        yaml_ast_walk(v, check.not_none(n.name))
        for value0 in n.values:
            yaml_ast_walk(v, value0)
    if isinstance(n, TagYamlNode):
        yaml_ast_walk_comment(v, n)
        yaml_ast_walk(v, check.not_none(n.value))
    if isinstance(n, DocumentYamlNode):
        yaml_ast_walk_comment(v, n)
        yaml_ast_walk(v, check.not_none(n.body))
    if isinstance(n, MappingYamlNode):
        yaml_ast_walk_comment(v, n)
        for value1 in n.values:
            yaml_ast_walk(v, value1)
    if isinstance(n, MappingKeyYamlNode):
        yaml_ast_walk_comment(v, n)
        yaml_ast_walk(v, check.not_none(n.value))
    if isinstance(n, MappingValueYamlNode):
        yaml_ast_walk_comment(v, n)
        yaml_ast_walk(v, n.key)
        yaml_ast_walk(v, n.value)
    if isinstance(n, SequenceYamlNode):
        yaml_ast_walk_comment(v, n)
        for value2 in n.values:
            yaml_ast_walk(v, check.not_none(value2))
    if isinstance(n, AnchorYamlNode):
        yaml_ast_walk_comment(v, n)
        yaml_ast_walk(v, check.not_none(n.name))
        yaml_ast_walk(v, check.not_none(n.value))
    if isinstance(n, AliasYamlNode):
        yaml_ast_walk_comment(v, n)
        yaml_ast_walk(v, check.not_none(n.value))


def yaml_ast_walk_comment(v: YamlAstVisitor, base: ta.Optional[BaseYamlNode]) -> None:
    if base is None:
        return
    if base.comment is None:
        return
    yaml_ast_walk(v, base.comment)


#


@dc.dataclass()
class FilterYamlAstWalker(YamlAstVisitor):
    typ: YamlNodeType = dc.field(default_factory=dataclass_field_required('typ'))
    results: ta.List[YamlNode] = dc.field(default_factory=list)

    def visit(self, n: YamlNode) -> YamlAstVisitor:
        if self.typ == n.type():
            self.results.append(n)
        return self


#


@dc.dataclass()
class YamlParentFinder:
    target: YamlNode

    def walk(self, parent: YamlNode, node: ta.Optional[YamlNode]) -> ta.Optional[YamlNode]:
        if self.target == node:
            return parent

        n = node
        if isinstance(n, CommentYamlNode):
            return None
        if isinstance(n, NullYamlNode):
            return None
        if isinstance(n, IntegerYamlNode):
            return None
        if isinstance(n, FloatYamlNode):
            return None
        if isinstance(n, StringYamlNode):
            return None
        if isinstance(n, MergeKeyYamlNode):
            return None
        if isinstance(n, BoolYamlNode):
            return None
        if isinstance(n, InfinityYamlNode):
            return None
        if isinstance(n, NanYamlNode):
            return None
        if isinstance(n, LiteralYamlNode):
            return self.walk(n, n.value)
        if isinstance(n, DirectiveYamlNode):
            if (found := self.walk(n, n.name)) is not None:
                return found
            for value0 in n.values:
                if (found := self.walk(n, value0)) is not None:
                    return found
        if isinstance(n, TagYamlNode):
            return self.walk(n, n.value)
        if isinstance(n, DocumentYamlNode):
            return self.walk(n, n.body)
        if isinstance(n, MappingYamlNode):
            for value1 in n.values:
                if (found := self.walk(n, value1)) is not None:
                    return found
        if isinstance(n, MappingKeyYamlNode):
            return self.walk(n, n.value)
        if isinstance(n, MappingValueYamlNode):
            if (found := self.walk(n, n.key)) is not None:
                return found
            return self.walk(n, n.value)
        if isinstance(n, SequenceYamlNode):
            for value2 in n.values:
                if (found := self.walk(n, value2)) is not None:
                    return found
        if isinstance(n, AnchorYamlNode):
            if (found := self.walk(n, n.name)) is not None:
                return found
            return self.walk(n, n.value)
        if isinstance(n, AliasYamlNode):
            return self.walk(n, n.value)
        return None


# Parent get parent node from child node.
def yaml_parent(root: YamlNode, child: YamlNode) -> ta.Optional[YamlNode]:
    finder = YamlParentFinder(target=child)
    return finder.walk(root, root)


#


# Filter returns a list of nodes that match the given type.
def yaml_filter(typ: YamlNodeType, node: YamlNode) -> ta.List[YamlNode]:
    walker = FilterYamlAstWalker(typ=typ)
    yaml_ast_walk(walker, node)
    return walker.results


# FilterFile returns a list of nodes that match the given type.
def yaml_filter_file(typ: YamlNodeType, file: YamlFile) -> ta.List[YamlNode]:
    results: ta.List[YamlNode] = []
    for doc in file.docs:
        walker = FilterYamlAstWalker(typ=typ)
        yaml_ast_walk(walker, doc)
        results.extend(walker.results)
    return results


#


@dc.dataclass()
class InvalidMergeTypeYamlError(YamlError):
    dst: YamlNode
    src: YamlNode

    @property
    def message(self) -> str:
        return f'cannot merge {self.src.type()} into {self.dst.type()}'


# Merge merge document, map, sequence node.
def yaml_ast_merge(dst: YamlNode, src: YamlNode) -> ta.Optional[YamlError]:
    if isinstance(src, DocumentYamlNode):
        doc: DocumentYamlNode = src
        src = check.not_none(doc.body)

    err = InvalidMergeTypeYamlError(dst=dst, src=src)
    if dst.type() == YamlNodeType.DOCUMENT:
        node0: DocumentYamlNode = check.isinstance(dst, DocumentYamlNode)
        return yaml_ast_merge(check.not_none(node0.body), src)
    if dst.type() == YamlNodeType.MAPPING:
        node1: MappingYamlNode = check.isinstance(dst, MappingYamlNode)
        if not isinstance(src, MappingYamlNode):
            return err
        target0: MappingYamlNode = src
        node1.merge(target0)
        return None
    if dst.type() == YamlNodeType.SEQUENCE:
        node2: SequenceYamlNode = check.isinstance(dst, SequenceYamlNode)
        if not isinstance(src, SequenceYamlNode):
            return err
        target1: SequenceYamlNode = src
        node2.merge(target1)
        return None
    return err



"""
TODO:
 - inheritance
 - dynamic registration
 - dynamic switching (skip docker if not running, skip online if not online, ...)
 - probably make IF_SINGLE understand parametErization
"""
import dataclasses as dc
import typing as ta

import pytest

from ..... import check
from ..... import collections as col
from .._registry import register
from .switches import SWITCHES
from .switches import SWITCHES_BY_NAME
from .switches import Switch
from .switches import SwitchState


Configable: ta.TypeAlias = pytest.FixtureRequest | pytest.Config


##


SWITCH_STATE_OPT_PREFIXES: ta.Mapping[SwitchState, str] = {
    SwitchState.ENABLED: '--',
    SwitchState.DISABLED: '--no-',
    SwitchState.ONLY: '--only-',
}


def _get_obj_config(obj: Configable) -> pytest.Config:
    if isinstance(obj, pytest.Config):
        return obj
    elif isinstance(obj, pytest.FixtureRequest):
        return obj.config
    else:
        raise TypeError(obj)


def is_disabled(obj: Configable | None, name: str) -> bool:
    check.isinstance(name, str)
    check.in_(name, SWITCHES_BY_NAME)
    return obj is not None and _get_obj_config(obj).getoption(f'--no-{name}')


def skip_if_disabled(obj: Configable | None, name: str) -> None:
    if is_disabled(obj, name):
        pytest.skip(f'{name} disabled')


def get_specified_switches(obj: Configable) -> dict[Switch, SwitchState]:
    ret: dict[Switch, SwitchState] = {}
    for sw in SWITCHES:
        sts = {
            st
            for st, pfx in SWITCH_STATE_OPT_PREFIXES.items()
            if _get_obj_config(obj).getoption(pfx + sw.name)
        }
        if sts:
            if len(sts) > 1:
                raise Exception(f'Multiple switches specified for {sw.name}')
            ret[sw] = check.single(sts)
    return ret


@dc.dataclass(frozen=True)
class ItemSwitches:
    switches: frozenset[Switch]
    not_switches: frozenset[Switch]


def get_item_switches(item: pytest.Item) -> ItemSwitches:
    return ItemSwitches(
        frozenset(sw for sw in SWITCHES if sw.attr in item.keywords),
        frozenset(sw for sw in SWITCHES if ('not_' + sw.attr) in item.keywords),
    )


@register
class SwitchesPlugin:
    def pytest_addoption(self, parser):
        for sw in SWITCHES:
            parser.addoption(f'--no-{sw.name}', action='store_true', default=False, help=f'disable {sw.name} tests')
            parser.addoption(f'--{sw.name}', action='store_true', default=False, help=f'enables {sw.name} tests')
            parser.addoption(f'--only-{sw.name}', action='store_true', default=False, help=f'enables only {sw.name} tests')  # noqa

    def pytest_configure(self, config):
        for sw in SWITCHES:
            config.addinivalue_line('markers', f'{sw.attr}: mark test as {sw.attr}')
            config.addinivalue_line('markers', f'not_{sw.attr}: mark test as not {sw.attr}')

    class _States:
        def __init__(self, session: pytest.Session) -> None:
            super().__init__()

            self.session = session

            self.default_states_by_switch: dict[Switch, SwitchState] = {
                sw: sw.default_state(session)
                for sw in SWITCHES
            }

            self.states_by_switch: dict[Switch, SwitchState] = {
                **self.default_states_by_switch,
                **get_specified_switches(session.config),
            }

            self.switch_sets_by_state: dict[SwitchState, frozenset[Switch]] = {
                st: frozenset(sws)
                for st, sws in col.multi_map(
                    (st, sw)
                    for sw, st in self.states_by_switch.items()
                ).items()
            }

    _states_key: ta.ClassVar[pytest.StashKey[_States]] = pytest.StashKey[_States]()

    def pytest_sessionstart(self, session):
        session.stash[self._states_key] = self._States(session)

    @pytest.hookimpl(tryfirst=True)
    def pytest_collection_modifyitems(self, config, items):
        def process_item(item):
            state: SwitchesPlugin._States = item.session.stash[self._states_key]

            item_switches = get_item_switches(item)

            for sw in item_switches.switches:
                for mk in sw.add_marks or []:
                    item.add_marker(mk)

            if (only_switches := state.switch_sets_by_state.get(SwitchState.ONLY)) is not None:
                if not any(sw in only_switches for sw in item_switches.switches):
                    item.add_marker(pytest.mark.skip(
                        reason=f'skipping switches ({tuple(sw.name for sw in item_switches.switches)!r} (only)',
                    ))
                    return

            disabled_switches = state.switch_sets_by_state.get(SwitchState.DISABLED, frozenset())
            for sw in item_switches.switches:
                if sw in disabled_switches:
                    item.add_marker(pytest.mark.skip(reason=f'skipping switch {sw.name!r} (disabled)'))

            enabled_switches = state.switch_sets_by_state.get(SwitchState.ENABLED, frozenset())
            for sw in item_switches.not_switches:
                if sw in enabled_switches:
                    item.add_marker(pytest.mark.skip(reason=f'skipping switch {sw.name!r} (not-enabled)'))

        for item in items:
            process_item(item)

    def pytest_collection_finish(self, session):
        state: SwitchesPlugin._States = session.stash[self._states_key]

        if_single_switches = state.switch_sets_by_state.get(SwitchState.IF_SINGLE, frozenset())

        if len(session.items) > 1:
            for item in session.items:
                item_switches = get_item_switches(item)
                for sw in item_switches.switches:
                    if sw in if_single_switches:
                        item.add_marker(pytest.mark.skip(reason=f'skipping switch {sw.name!r} (not-single)'))



# ruff: noqa: UP006 UP007 UP045
import typing as ta

from .proxy import AsyncIoProxy
from .proxy import _register_async_io_proxy_cls


SelfT = ta.TypeVar('SelfT')

AnyStrT = ta.TypeVar('AnyStrT', bytes, str)


##


@_register_async_io_proxy_cls
class TypingIO_AsyncIoProxy(AsyncIoProxy, ta.Generic[AnyStrT], proxied_cls=ta.IO):  # noqa
    @property
    def mode(self) -> str:
        raise TypeError

    @property
    def name(self) -> str:
        raise TypeError

    async def close(self) -> None:
        raise TypeError

    @property
    def closed(self) -> bool:
        raise TypeError

    def fileno(self) -> int:
        raise TypeError

    async def flush(self) -> None:
        raise TypeError

    def isatty(self) -> bool:
        raise TypeError

    async def read(self, n: int = -1) -> AnyStrT:
        raise TypeError

    def readable(self) -> bool:
        raise TypeError

    async def readline(self, limit: int = -1) -> AnyStrT:
        raise TypeError

    async def readlines(self, hint: int = -1) -> ta.List[AnyStrT]:
        raise TypeError

    async def seek(self, offset: int, whence: int = 0) -> int:
        raise TypeError

    def seekable(self) -> bool:
        raise TypeError

    async def tell(self) -> int:
        raise TypeError

    async def truncate(self, size: ta.Optional[int] = None) -> int:
        raise TypeError

    def writable(self) -> bool:
        raise TypeError

    async def write(self, s: AnyStrT) -> int:
        raise TypeError

    async def writelines(self, lines: ta.List[AnyStrT]) -> None:
        raise TypeError

    async def __aenter__(self: SelfT) -> SelfT:
        raise TypeError

    async def __aexit__(self, exc_type, exc_value, exc_tb):
        raise TypeError


@_register_async_io_proxy_cls
class TypingBinaryIO_AsyncIoProxy(TypingIO_AsyncIoProxy[bytes], proxied_cls=ta.BinaryIO):  # noqa
    def write(self, s: ta.Union[bytes, bytearray]) -> int:  # type: ignore[override]
        raise TypeError


@_register_async_io_proxy_cls
class TypingTextIO_AsyncIoProxy(TypingIO_AsyncIoProxy[str], proxied_cls=ta.TextIO):  # noqa
    # @property
    # def buffer(self) -> BinaryIO:
    #     pass

    @property
    def encoding(self) -> str:
        raise TypeError

    @property
    def errors(self) -> ta.Optional[str]:
        raise TypeError

    @property
    def line_buffering(self) -> bool:
        raise TypeError

    @property
    def newlines(self) -> ta.Any:
        raise TypeError



import abc
import collections.abc
import typing as ta

from ... import check
from ... import dataclasses as dc
from ... import lang
from ... import reflect as rfl
from ..api.contexts import UnmarshalContext
from ..api.contexts import UnmarshalFactoryContext
from ..api.types import Unmarshaler
from ..api.types import UnmarshalerFactory
from ..api.values import Value
from .api import FieldTypeTagging
from .api import Impls
from .api import Polymorphism
from .api import TypeTagging
from .api import WrapperTypeTagging
from .impls import get_polymorphism_impls


##


class PolymorphismUnmarshaler(Unmarshaler, lang.Abstract):
    @abc.abstractmethod
    def get_impls(self) -> ta.Mapping[str, Unmarshaler]:
        raise NotImplementedError


@dc.dataclass(frozen=True)
class WrapperPolymorphismUnmarshaler(PolymorphismUnmarshaler):
    m: ta.Mapping[str, Unmarshaler]

    def get_impls(self) -> ta.Mapping[str, Unmarshaler]:
        return self.m

    def unmarshal(self, ctx: UnmarshalContext, v: Value) -> ta.Any | None:
        ma = check.isinstance(v, collections.abc.Mapping)
        [(tag, iv)] = ma.items()
        u = self.m[tag]
        return u.unmarshal(ctx, iv)


@dc.dataclass(frozen=True)
class FieldPolymorphismUnmarshaler(PolymorphismUnmarshaler):
    m: ta.Mapping[str, Unmarshaler]
    tf: str

    def get_impls(self) -> ta.Mapping[str, Unmarshaler]:
        return self.m

    def unmarshal(self, ctx: UnmarshalContext, v: Value) -> ta.Any | None:
        ma = dict(check.isinstance(v, collections.abc.Mapping))
        tag = ma.pop(self.tf)
        u = self.m[tag]
        return u.unmarshal(ctx, ma)


def make_polymorphism_unmarshaler(
        impls: Impls,
        tt: TypeTagging,
        ctx: UnmarshalFactoryContext,
) -> Unmarshaler:
    check.not_empty(impls)

    m = {
        t: u
        for i in impls
        for u in [ctx.make_unmarshaler(i.ty)]
        for t in [i.tag, *i.alts]
    }

    if isinstance(tt, WrapperTypeTagging):
        return WrapperPolymorphismUnmarshaler(m)
    elif isinstance(tt, FieldTypeTagging):
        return FieldPolymorphismUnmarshaler(m, tt.field)
    else:
        raise TypeError(tt)


@dc.dataclass(frozen=True)
class PolymorphismUnmarshalerFactory(UnmarshalerFactory):
    p: Polymorphism
    tt: TypeTagging = WrapperTypeTagging()

    def make_unmarshaler(self, ctx: UnmarshalFactoryContext, rty: rfl.Type) -> ta.Callable[[], Unmarshaler] | None:
        if (impls := get_polymorphism_impls(rty, self.p)) is None:
            return None
        return lambda: make_polymorphism_unmarshaler(impls, self.tt, ctx)



from .. import lang
from .. import marshal as msh
from .base import Bootstrap
from .harness import BOOTSTRAP_TYPES_BY_NAME


##


@lang.static_init
def _install_standard_marshaling() -> None:
    cfgs_poly = msh.Polymorphism(
        Bootstrap.Config,
        [msh.Impl(b.Config, n) for n, b in BOOTSTRAP_TYPES_BY_NAME.items()],
    )
    msh.install_standard_factories(
        msh.PolymorphismMarshalerFactory(cfgs_poly),
        msh.PolymorphismUnmarshalerFactory(cfgs_poly),
    )
